#
# Write a symbol to python
#
from __future__ import annotations

import numbers
from collections import OrderedDict

import numpy as np
import scipy.sparse
from numpy.typing import ArrayLike

import pybamm

if pybamm.has_jax():
    import jax

    platform = jax.lib.xla_bridge.get_backend().platform.casefold()
    if platform != "metal":
        jax.config.update("jax_enable_x64", True)


class JaxCooMatrix:
    """
    A sparse matrix in COO format, with internal arrays using jax device arrays

    This matrix only has two operations supported, a multiply with a scalar, and a
    dot product with a dense vector. It can also be converted to a dense 2D jax
    device array

    Parameters
    ----------

    row: arraylike
        1D array holding row indices of non-zero entries
    col: arraylike
        1D array holding col indices of non-zero entries
    data: arraylike
        1D array holding non-zero entries
    shape: 2-element tuple (x, y)
        where x is the number of rows, and y the number of columns of the matrix
    """

    def __init__(
        self, row: ArrayLike, col: ArrayLike, data: ArrayLike, shape: tuple[int, int]
    ):
        if not pybamm.has_jax():  # pragma: no cover
            raise ModuleNotFoundError(
                "Jax or jaxlib is not installed, please see https://docs.pybamm.org/en/latest/source/user_guide/installation/gnu-linux-mac.html#optional-jaxsolver"
            )

        self.row = jax.numpy.array(row)
        self.col = jax.numpy.array(col)
        self.data = jax.numpy.array(data)
        self.shape = shape
        self.nnz = len(self.data)

    def toarray(self):
        """convert sparse matrix to a dense 2D array"""
        result = jax.numpy.zeros(self.shape, dtype=self.data.dtype)
        return result.at[self.row, self.col].add(self.data)

    def dot_product(self, b):
        """
        dot product of matrix with a dense column vector b

        Parameters
        ----------
        b: jax device array
            must have shape (n, 1)
        """
        # assume b is a column vector
        result = jax.numpy.zeros((self.shape[0], 1), dtype=b.dtype)
        return result.at[self.row].add(self.data.reshape(-1, 1) * b[self.col])

    def scalar_multiply(self, b: float):
        """
        multiply of matrix with a scalar b

        Parameters
        ----------
        b: Number or 1 element jax device array
            scalar value to multiply
        """
        # assume b is a scalar or ndarray with 1 element
        return JaxCooMatrix(self.row, self.col, (self.data * b).reshape(-1), self.shape)

    def multiply(self, b):
        """
        general matrix multiply not supported
        """
        raise NotImplementedError

    def __matmul__(self, b):
        """see self.dot_product"""
        return self.dot_product(b)


def create_jax_coo_matrix(value: scipy.sparse):
    """
    Creates a JaxCooMatrix from a scipy.sparse matrix

    Parameters
    ----------

    value: scipy.sparse matrix
        the sparse matrix to be converted
    """
    scipy_coo = value.tocoo()
    row = jax.numpy.asarray(scipy_coo.row)
    col = jax.numpy.asarray(scipy_coo.col)
    data = jax.numpy.asarray(scipy_coo.data)
    return JaxCooMatrix(row, col, data, value.shape)


def id_to_python_variable(symbol_id, constant=False):
    """
    This function defines the format for the python variable names used in find_symbols
    and to_python. Variable names are based on a nodes' id to make them unique
    """

    if constant:
        var_format = "const_{:05d}"
    else:
        var_format = "var_{:05d}"

    # Need to replace "-" character to make them valid python variable names
    return var_format.format(symbol_id).replace("-", "m")


def is_scalar(arg):
    is_number = isinstance(arg, numbers.Number)
    if is_number:
        return True
    else:
        return np.all(np.array(arg.shape) == 1)


def find_symbols(
    symbol: pybamm.Symbol,
    constant_symbols: OrderedDict,
    variable_symbols: OrderedDict,
    output_jax=False,
):
    """
    This function converts an expression tree to a dictionary of node id's and strings
    specifying valid python code to calculate that nodes value, given y and t.

    The function distinguishes between nodes that represent constant nodes in the tree
    (e.g. a pybamm.Matrix), and those that are variable (e.g. subtrees that contain
    pybamm.StateVector). The former are put in `constant_symbols`, the latter in
    `variable_symbols`

    Note that it is important that the arguments `constant_symbols` and
    `variable_symbols` be an *ordered* dict, since the final ordering of the code lines
    are important for the calculations. A dict is specified rather than a list so that
    identical subtrees (which give identical id's) are not recalculated in the code

    Parameters
    ----------
    symbol : :class:`pybamm.Symbol`
        The symbol or expression tree to convert

    constant_symbol: collections.OrderedDict
        The output dictionary of constant symbol ids to lines of code

    variable_symbol: collections.OrderedDict
        The output dictionary of variable (with y or t) symbol ids to lines of code

    output_jax: bool
        If True, only numpy and jax operations will be used in the generated code,
        raises NotImplNotImplementedError if any SparseStack or Mat-Mat multiply
        operations are used

    """
    # constant symbols that are not numbers are stored in a list of constants, which are
    # passed into the generated function constant symbols that are numbers are written
    # directly into the code
    if symbol.is_constant():
        value = symbol.evaluate()
        if not isinstance(value, numbers.Number):
            if output_jax and scipy.sparse.issparse(value):
                # convert any remaining sparse matrices to our custom coo matrix
                constant_symbols[symbol.id] = create_jax_coo_matrix(value)
            else:
                constant_symbols[symbol.id] = value
        return

    # process children recursively
    for child in symbol.children:
        find_symbols(child, constant_symbols, variable_symbols, output_jax)

    # calculate the variable names that will hold the result of calculating the
    # children variables
    children_vars = []
    for child in symbol.children:
        if child.is_constant():
            child_eval = child.evaluate()
            if isinstance(child_eval, numbers.Number):
                children_vars.append(str(child_eval))
            else:
                children_vars.append(id_to_python_variable(child.id, True))
        else:
            children_vars.append(id_to_python_variable(child.id, False))

    if isinstance(symbol, pybamm.BinaryOperator):
        # Multiplication and Division need special handling for scipy sparse matrices
        # TODO: we can pass through a dummy y and t to get the type and then hardcode
        # the right line, avoiding these checks
        if isinstance(symbol, pybamm.Multiplication):
            dummy_eval_left = symbol.children[0].evaluate_for_shape()
            dummy_eval_right = symbol.children[1].evaluate_for_shape()
            if scipy.sparse.issparse(dummy_eval_left):
                if output_jax and is_scalar(dummy_eval_right):
                    symbol_str = (
                        f"{children_vars[0]}.scalar_multiply({children_vars[1]})"
                    )
                else:
                    symbol_str = f"{children_vars[0]}.multiply({children_vars[1]})"
            elif scipy.sparse.issparse(dummy_eval_right):
                symbol_str = f"{children_vars[1]}.multiply({children_vars[0]})"
            else:
                symbol_str = f"{children_vars[0]} * {children_vars[1]}"
        elif isinstance(symbol, pybamm.Division):
            dummy_eval_left = symbol.children[0].evaluate_for_shape()
            dummy_eval_right = symbol.children[1].evaluate_for_shape()
            if scipy.sparse.issparse(dummy_eval_left):
                if output_jax and is_scalar(dummy_eval_right):
                    symbol_str = (
                        f"{children_vars[0]}.scalar_multiply(1/{children_vars[1]})"
                    )
                else:
                    symbol_str = f"{children_vars[0]}.multiply(1/{children_vars[1]})"
            else:
                symbol_str = f"{children_vars[0]} / {children_vars[1]}"

        elif isinstance(symbol, pybamm.Inner):
            dummy_eval_left = symbol.children[0].evaluate_for_shape()
            dummy_eval_right = symbol.children[1].evaluate_for_shape()
            if scipy.sparse.issparse(dummy_eval_left):
                if output_jax and is_scalar(dummy_eval_right):
                    symbol_str = (
                        f"{children_vars[0]}.scalar_multiply({children_vars[1]})"
                    )
                else:
                    symbol_str = f"{children_vars[0]}.multiply({children_vars[1]})"
            elif scipy.sparse.issparse(dummy_eval_right):
                if output_jax and is_scalar(dummy_eval_left):
                    symbol_str = (
                        f"{children_vars[1]}.scalar_multiply({children_vars[0]})"
                    )
                else:
                    symbol_str = f"{children_vars[1]}.multiply({children_vars[0]})"
            else:
                symbol_str = f"{children_vars[0]} * {children_vars[1]}"

        elif isinstance(symbol, pybamm.Minimum):
            symbol_str = f"np.minimum({children_vars[0]},{children_vars[1]})"
        elif isinstance(symbol, pybamm.Maximum):
            symbol_str = f"np.maximum({children_vars[0]},{children_vars[1]})"

        elif isinstance(symbol, pybamm.MatrixMultiplication):
            dummy_eval_left = symbol.children[0].evaluate_for_shape()
            dummy_eval_right = symbol.children[1].evaluate_for_shape()
            if output_jax and (
                scipy.sparse.issparse(dummy_eval_left)
                and scipy.sparse.issparse(dummy_eval_right)
            ):
                raise NotImplementedError(
                    "sparse mat-mat multiplication not supported for output_jax == True"
                )
            else:
                symbol_str = (
                    children_vars[0] + " " + symbol.name + " " + children_vars[1]
                )
        else:
            symbol_str = children_vars[0] + " " + symbol.name + " " + children_vars[1]

    elif isinstance(symbol, pybamm.UnaryOperator):
        # Index has a different syntax than other univariate operations
        if isinstance(symbol, pybamm.Index):
            symbol_str = f"{children_vars[0]}[{symbol.slice.start}:{symbol.slice.stop}]"
        else:
            symbol_str = symbol.name + "(" + children_vars[0] + ")"

    elif isinstance(symbol, pybamm.Function):
        children_str = ""
        for child_var in children_vars:
            if children_str == "":
                children_str = child_var
            else:
                children_str += ", " + child_var
        if isinstance(symbol.function, np.ufunc):
            # write any numpy functions directly
            symbol_str = f"np.{symbol.function.__name__}({children_str})"
        else:
            # unknown function, store it as a constant and call this in the
            # generated code
            constant_symbols[symbol.id] = symbol.function
            funct_var = id_to_python_variable(symbol.id, True)
            symbol_str = f"{funct_var}({children_str})"

    elif isinstance(symbol, pybamm.Concatenation):
        # no need to concatenate if there is only a single child
        if isinstance(symbol, pybamm.NumpyConcatenation):
            if len(children_vars) == 1:
                symbol_str = children_vars[0]
            else:
                symbol_str = "np.concatenate(({}))".format(",".join(children_vars))

        elif isinstance(symbol, pybamm.SparseStack):
            if len(children_vars) == 1:
                symbol_str = children_vars[0]
            else:
                if output_jax:
                    raise NotImplementedError
                else:
                    symbol_str = "scipy.sparse.vstack(({}))".format(
                        ",".join(children_vars)
                    )

        # DomainConcatenation specifies a particular ordering for the concatenation,
        # which we must follow
        elif isinstance(symbol, pybamm.DomainConcatenation):
            slice_starts = []
            all_child_vectors = []
            for i in range(symbol.secondary_dimensions_npts):
                child_vectors = []
                for child_var, slices in zip(
                    children_vars, symbol._children_slices, strict=False
                ):
                    for child_dom, child_slice in slices.items():
                        slice_starts.append(symbol._slices[child_dom][i].start)
                        child_vectors.append(
                            f"{child_var}[{child_slice[i].start}:{child_slice[i].stop}]"
                        )
                all_child_vectors.extend(
                    [
                        v
                        for _, v in sorted(
                            zip(slice_starts, child_vectors, strict=False)
                        )
                    ]
                )
            if len(children_vars) > 1 or symbol.secondary_dimensions_npts > 1:
                symbol_str = "np.concatenate(({}))".format(",".join(all_child_vectors))
            else:
                symbol_str = "{}".format(",".join(children_vars))
        else:
            raise NotImplementedError

    # Note: we assume that y is being passed as a column vector
    elif isinstance(symbol, pybamm.StateVector):
        indices = np.argwhere(symbol.evaluation_array).reshape(-1).astype(np.int32)
        consecutive = np.all(indices[1:] - indices[:-1] == 1)
        if len(indices) == 1 or consecutive:
            symbol_str = f"y[{indices[0]}:{indices[-1] + 1}]"
        else:
            indices_array = pybamm.Array(indices)
            constant_symbols[indices_array.id] = indices
            index_name = id_to_python_variable(indices_array.id, True)
            symbol_str = f"y[{index_name}]"

    elif isinstance(symbol, pybamm.Time):
        symbol_str = "t"

    elif isinstance(symbol, pybamm.InputParameter):
        symbol_str = f'inputs["{symbol.name}"]'

    else:
        raise NotImplementedError(
            f"Conversion to python not implemented for a symbol of type '{type(symbol)}'"
        )

    variable_symbols[symbol.id] = symbol_str


def to_python(
    symbol: pybamm.Symbol, debug=False, output_jax=False
) -> tuple[OrderedDict, str]:
    """
    This function converts an expression tree into a dict of constant input values, and
    valid python code that acts like the tree's :func:`pybamm.Symbol.evaluate` function

    Parameters
    ----------
    symbol : :class:`pybamm.Symbol`
        The symbol to convert to python code

    debug : bool
        If set to True, the function also emits debug code

    Returns
    -------
    collections.OrderedDict:
        dict mapping node id to a constant value. Represents all the constant nodes in
        the expression tree
    str:
        valid python code that will evaluate all the variable nodes in the tree.
    output_jax: bool
        If True, only numpy and jax operations will be used in the generated code.
        Raises NotImplNotImplementedError if any SparseStack or Mat-Mat multiply
        operations are used

    """
    constant_values: OrderedDict = OrderedDict()
    variable_symbols: OrderedDict = OrderedDict()
    find_symbols(symbol, constant_values, variable_symbols, output_jax)

    line_format = "{} = {}"

    if debug:  # pragma: no cover
        variable_lines = [
            f"print('{line_format.format(id_to_python_variable(symbol_id, False), symbol_line)}'); "
            + line_format.format(id_to_python_variable(symbol_id, False), symbol_line)
            + "; print(type({0}),np.shape({0}))".format(
                id_to_python_variable(symbol_id, False)
            )
            for symbol_id, symbol_line in variable_symbols.items()
        ]
    else:
        variable_lines = [
            line_format.format(id_to_python_variable(symbol_id, False), symbol_line)
            for symbol_id, symbol_line in variable_symbols.items()
        ]

    return constant_values, "\n".join(variable_lines)


class EvaluatorPython:
    """
    Converts a pybamm expression tree into pure python code that will calculate the
    result of calling `evaluate(t, y)` on the given expression tree.

    Parameters
    ----------

    symbol : :class:`pybamm.Symbol`
        The symbol to convert to python code


    """

    def __init__(self, symbol: pybamm.Symbol):
        constants, python_str = pybamm.to_python(symbol, debug=False)

        # extract constants in generated function
        for i, symbol_id in enumerate(constants.keys()):
            const_name = id_to_python_variable(symbol_id, True)
            python_str = f"{const_name} = constants[{i}]\n" + python_str

        # constants passed in as an ordered dict, convert to list
        self._constants = list(constants.values())

        # indent code
        python_str = "   " + python_str
        python_str = python_str.replace("\n", "\n   ")

        # add function def to first line
        python_str = (
            "def evaluate(constants, t=None, y=None, inputs=None):\n" + python_str
        )

        # calculate the final variable that will output the result of calling `evaluate`
        # on `symbol`
        result_var = id_to_python_variable(symbol.id, symbol.is_constant())
        if symbol.is_constant():
            result_value = symbol.evaluate()

        # add return line
        if symbol.is_constant() and isinstance(result_value, numbers.Number):
            python_str = python_str + "\n   return " + str(result_value)
        else:
            python_str = python_str + "\n   return " + result_var

        # store a copy of examine_jaxpr
        python_str = python_str + "\nself._evaluate = evaluate"

        self._python_str = python_str
        self._result_var = result_var
        self._symbol = symbol

        # compile and run the generated python code,
        compiled_function = compile(python_str, result_var, "exec")
        exec(compiled_function)

    def __call__(self, t=None, y=None, inputs=None):
        """
        evaluate function
        """
        # generated code assumes y is a column vector
        if y is not None and y.ndim == 1:
            y = y.reshape(-1, 1)

        result = self._evaluate(self._constants, t, y, inputs)

        return result

    def __getstate__(self):
        # Control the state of instances of EvaluatorPython
        # before pickling. Method "_evaluate" cannot be pickled.
        # See https://github.com/pybamm-team/PyBaMM/issues/1283
        state = self.__dict__.copy()
        del state["_evaluate"]
        return state

    def __setstate__(self, state):
        # Restore pickled attributes and
        # compile code from "python_str"
        # Execution of bytecode (re)adds attribute
        # "_method"
        self.__dict__.update(state)
        compiled_function = compile(self._python_str, self._result_var, "exec")
        exec(compiled_function)


class EvaluatorJax:
    """
    Converts a pybamm expression tree into pure python code that will calculate the
    result of calling `evaluate(t, y)` on the given expression tree. The resultant code
    is compiled with JAX

    Limitations: JAX currently does not work on expressions involving sparse matrices,
    so any sparse matrices and operations involved sparse matrices are converted to
    their dense equivilents before compilation

    Parameters
    ----------

    symbol : :class:`pybamm.Symbol`
        The symbol to convert to python code


    """

    def __init__(self, symbol: pybamm.Symbol):
        if not pybamm.has_jax():  # pragma: no cover
            raise ModuleNotFoundError(
                "Jax or jaxlib is not installed, please see https://docs.pybamm.org/en/latest/source/user_guide/installation/gnu-linux-mac.html#optional-jaxsolver"
            )

        constants, python_str = pybamm.to_python(symbol, debug=False, output_jax=True)

        # replace numpy function calls to jax numpy calls
        python_str = python_str.replace("np.", "jax.numpy.")

        # convert all numpy constants to device vectors
        for symbol_id in constants:
            if isinstance(constants[symbol_id], np.ndarray):
                constants[symbol_id] = jax.device_put(constants[symbol_id])

        # get a list of constant arguments to input to the function
        self._arg_list = [
            id_to_python_variable(symbol_id, True) for symbol_id in constants.keys()
        ]

        # get a list of hashable arguments to make static
        # a jax device array is not hashable
        static_argnums = (
            i
            for i, c in enumerate(constants.values())
            if not (isinstance(c, jax.Array))
        )

        # store constants
        self._constants = tuple(constants.values())

        # indent code
        python_str = "   " + python_str
        python_str = python_str.replace("\n", "\n   ")

        # add function def to first line
        args = "t=None, y=None, inputs=None"
        if self._arg_list:
            args = ",".join(self._arg_list) + ", " + args
        python_str = f"def evaluate_jax({args}):\n" + python_str

        # calculate the final variable that will output the result of calling `evaluate`
        # on `symbol`
        result_var = id_to_python_variable(symbol.id, symbol.is_constant())
        if symbol.is_constant():
            result_value = symbol.evaluate()

        # add return line
        if symbol.is_constant() and isinstance(result_value, numbers.Number):
            python_str = python_str + "\n   return " + str(result_value)
        else:
            python_str = python_str + "\n   return " + result_var

        # store a copy of examine_jaxpr
        python_str = python_str + "\nself._evaluate_jax = evaluate_jax"

        # store the final generated code
        self._python_str = python_str

        # compile and run the generated python code,
        compiled_function = compile(python_str, result_var, "exec")
        exec(compiled_function)

        self._static_argnums = tuple(static_argnums)
        self._jit_evaluate = jax.jit(
            self._evaluate_jax,  # type:ignore[attr-defined]
            static_argnums=self._static_argnums,
        )

    @property
    def _constants(self):
        return self.__constants

    @_constants.setter
    def _constants(self, value):
        self.__constants = value

    def get_jacobian(self):
        n = len(self._arg_list)

        # forward mode autodiff  wrt y, which is argument 1 after arg_list
        jacobian_evaluate = jax.jacfwd(self._evaluate_jax, argnums=1 + n)

        self._jac_evaluate = jax.jit(
            jacobian_evaluate, static_argnums=self._static_argnums
        )

        return EvaluatorJaxJacobian(self._jac_evaluate, self._constants)

    def get_jacobian_action(self):
        return self.jvp

    def get_sensitivities(self):
        n = len(self._arg_list)

        # forward mode autodiff wrt inputs, which is argument 2 after arg_list
        jacobian_evaluate = jax.jacfwd(self._evaluate_jax, argnums=2 + n)

        self._sens_evaluate = jax.jit(
            jacobian_evaluate, static_argnums=self._static_argnums
        )

        return EvaluatorJaxSensitivities(self._sens_evaluate, self._constants)

    def debug(self, t=None, y=None, inputs=None):
        # generated code assumes y is a column vector
        if y is not None and y.ndim == 1:
            y = y.reshape(-1, 1)

        # execute code
        jaxpr = jax.make_jaxpr(self._evaluate_jax)(*self._constants, t, y, inputs).jaxpr
        print("invars:", jaxpr.invars)
        print("outvars:", jaxpr.outvars)
        print("constvars:", jaxpr.constvars)
        for eqn in jaxpr.eqns:
            print("equation:", eqn.invars, eqn.primitive, eqn.outvars, eqn.params)
        print()
        print("jaxpr:", jaxpr)

    def __call__(self, t=None, y=None, inputs=None):
        """
        evaluate function
        """
        # generated code assumes y is a column vector
        if y is not None and y.ndim == 1:
            y = y.reshape(-1, 1)

        result = self._jit_evaluate(*self._constants, t, y, inputs)

        return result

    def jvp(self, t=None, y=None, v=None, inputs=None):
        """
        evaluate jacobian vector product of function
        """

        # generated code assumes y is a column vector
        if y is not None and y.ndim == 1:
            y = y.reshape(-1, 1)
        if v is not None and v.ndim == 1:
            v = v.reshape(-1, 1)

        def bind_t_and_inputs(the_y):
            return self._jit_evaluate(*self._constants, t, the_y, inputs)

        return jax.jvp(bind_t_and_inputs, (y,), (v,))[1]


class EvaluatorJaxJacobian:
    def __init__(self, jac_evaluate, constants):
        self._jac_evaluate = jac_evaluate
        self._constants = constants

    def __call__(self, t=None, y=None, inputs=None):
        """
        evaluate function
        """
        # generated code assumes y is a column vector
        if y is not None and y.ndim == 1:
            y = y.reshape(-1, 1)

        # execute code
        result = self._jac_evaluate(*self._constants, t, y, inputs)
        result = result.reshape(result.shape[0], -1)

        return result


class EvaluatorJaxSensitivities:
    def __init__(self, jac_evaluate, constants):
        self._jac_evaluate = jac_evaluate
        self._constants = constants

    def __call__(self, t=None, y=None, inputs=None):
        """
        evaluate function
        """
        # generated code assumes y is a column vector
        if y is not None and y.ndim == 1:
            y = y.reshape(-1, 1)

        # execute code
        result = self._jac_evaluate(*self._constants, t, y, inputs)
        result = {
            key: value.reshape(value.shape[0], -1) for key, value in result.items()
        }

        return result



import pybamm

whole_cell = ["negative electrode", "separator", "positive electrode"]

# Domains at cell centres
x_n = pybamm.SpatialVariable(
    "x_n",
    domain=["negative electrode"],
    auxiliary_domains={"secondary": "current collector"},
    coord_sys="cartesian",
)
x_s = pybamm.SpatialVariable(
    "x_s",
    domain=["separator"],
    auxiliary_domains={"secondary": "current collector"},
    coord_sys="cartesian",
)
x_p = pybamm.SpatialVariable(
    "x_p",
    domain=["positive electrode"],
    auxiliary_domains={"secondary": "current collector"},
    coord_sys="cartesian",
)
x = pybamm.SpatialVariable(
    "x",
    domain=whole_cell,
    auxiliary_domains={"secondary": "current collector"},
    coord_sys="cartesian",
)

y = pybamm.SpatialVariable("y", domain="current collector", coord_sys="cartesian")
z = pybamm.SpatialVariable("z", domain="current collector", coord_sys="cartesian")
r_macro = pybamm.SpatialVariable(
    "r_macro", domain="current collector", coord_sys="cylindrical polar"
)

r_n = pybamm.SpatialVariable(
    "r_n",
    domain=["negative particle"],
    auxiliary_domains={
        "secondary": "negative electrode",
        "tertiary": "current collector",
    },
    coord_sys="spherical polar",
)
r_p = pybamm.SpatialVariable(
    "r_p",
    domain=["positive particle"],
    auxiliary_domains={
        "secondary": "positive electrode",
        "tertiary": "current collector",
    },
    coord_sys="spherical polar",
)
r_n_prim = pybamm.SpatialVariable(
    "r_n_prim",
    domain=["negative primary particle"],
    auxiliary_domains={
        "secondary": "negative electrode",
        "tertiary": "current collector",
    },
    coord_sys="spherical polar",
)
r_p_prim = pybamm.SpatialVariable(
    "r_p_prim",
    domain=["positive primary particle"],
    auxiliary_domains={
        "secondary": "positive electrode",
        "tertiary": "current collector",
    },
    coord_sys="spherical polar",
)
r_n_sec = pybamm.SpatialVariable(
    "r_n_sec",
    domain=["negative secondary particle"],
    auxiliary_domains={
        "secondary": "negative electrode",
        "tertiary": "current collector",
    },
    coord_sys="spherical polar",
)
r_p_sec = pybamm.SpatialVariable(
    "r_p_sec",
    domain=["positive secondary particle"],
    auxiliary_domains={
        "secondary": "positive electrode",
        "tertiary": "current collector",
    },
    coord_sys="spherical polar",
)

R_n = pybamm.SpatialVariable(
    "R_n",
    domain=["negative particle size"],
    auxiliary_domains={
        "secondary": "negative electrode",
        "tertiary": "current collector",
    },
    coord_sys="cartesian",
)

R_p = pybamm.SpatialVariable(
    "R_p",
    domain=["positive particle size"],
    auxiliary_domains={
        "secondary": "positive electrode",
        "tertiary": "current collector",
    },
    coord_sys="cartesian",
)

R_n_prim = pybamm.SpatialVariable(
    "R_n_prim",
    domain=["negative primary particle size"],
    auxiliary_domains={
        "secondary": "negative electrode",
        "tertiary": "current collector",
    },
    coord_sys="cartesian",
)
R_p_prim = pybamm.SpatialVariable(
    "R_p_prim",
    domain=["positive primary particle size"],
    auxiliary_domains={
        "secondary": "positive electrode",
        "tertiary": "current collector",
    },
    coord_sys="cartesian",
)

R_n_sec = pybamm.SpatialVariable(
    "R_n_sec",
    domain=["negative secondary particle size"],
    auxiliary_domains={
        "secondary": "negative electrode",
        "tertiary": "current collector",
    },
    coord_sys="cartesian",
)
R_p_sec = pybamm.SpatialVariable(
    "R_p_sec",
    domain=["positive secondary particle size"],
    auxiliary_domains={
        "secondary": "positive electrode",
        "tertiary": "current collector",
    },
    coord_sys="cartesian",
)

# Domains at cell edges
x_n_edge = pybamm.SpatialVariableEdge(
    "x_n",
    domain=["negative electrode"],
    auxiliary_domains={"secondary": "current collector"},
    coord_sys="cartesian",
)
x_s_edge = pybamm.SpatialVariableEdge(
    "x_s",
    domain=["separator"],
    auxiliary_domains={"secondary": "current collector"},
    coord_sys="cartesian",
)
x_p_edge = pybamm.SpatialVariableEdge(
    "x_p",
    domain=["positive electrode"],
    auxiliary_domains={"secondary": "current collector"},
    coord_sys="cartesian",
)
x_edge = pybamm.SpatialVariableEdge(
    "x",
    domain=whole_cell,
    auxiliary_domains={"secondary": "current collector"},
    coord_sys="cartesian",
)

y_edge = pybamm.SpatialVariableEdge(
    "y", domain="current collector", coord_sys="cartesian"
)
z_edge = pybamm.SpatialVariableEdge(
    "z", domain="current collector", coord_sys="cartesian"
)
r_macro_edge = pybamm.SpatialVariableEdge(
    "r_macro", domain="current collector", coord_sys="cylindrical polar"
)


r_n_edge = pybamm.SpatialVariableEdge(
    "r_n",
    domain=["negative particle"],
    auxiliary_domains={
        "secondary": "negative electrode",
        "tertiary": "current collector",
    },
    coord_sys="spherical polar",
)
r_p_edge = pybamm.SpatialVariableEdge(
    "r_p",
    domain=["positive particle"],
    auxiliary_domains={
        "secondary": "positive electrode",
        "tertiary": "current collector",
    },
    coord_sys="spherical polar",
)

R_n_edge = pybamm.SpatialVariableEdge(
    "R_n",
    domain=["negative particle size"],
    auxiliary_domains={
        "secondary": "negative electrode",
        "tertiary": "current collector",
    },
    coord_sys="cartesian",
)
R_p_edge = pybamm.SpatialVariableEdge(
    "R_p",
    domain=["positive particle size"],
    auxiliary_domains={
        "secondary": "positive electrode",
        "tertiary": "current collector",
    },
    coord_sys="cartesian",
)



#
# Newman Tobias Model
#
import pybamm

from .dfn import DFN


class NewmanTobias(DFN):
    """
    Newman-Tobias model of a lithium-ion battery based on the formulation in
    :footcite:t:`Newman1962`. This model assumes a uniform concentration profile in the
    electrolyte. Unlike the model posed in :footcite:t:`Newman1962`, this model accounts
    for nonlinear Butler-Volmer kinetics. It also tracks the average concentration in
    the solid phase in each electrode, which is equivalent to including an equation for
    the local state of charge as in :footcite:t:`Chu2020`. The user can pass the
    "particle" option to include mass transport in the particles.

    See :class:`pybamm.lithium_ion.BaseModel` for more details.

    """

    def __init__(self, options=None, name="Newman-Tobias model", build=True):
        # Set default option "uniform profile" for particle submodel. Other
        # default options are those given in `pybamm.BatteryModelOptions` defined in
        # `base_battery_model.py`.
        options = options or {}
        if "particle" not in options:
            options["particle"] = "uniform profile"
        self.x_average = True

        super().__init__(options, name, build)

        pybamm.citations.register("Newman1962")
        pybamm.citations.register("Chu2020")

    def set_particle_submodel(self):
        for domain in ["negative", "positive"]:
            particle = getattr(self.options, domain)["particle"]
            for phase in self.options.phases[domain]:
                if particle == "Fickian diffusion":
                    submod = pybamm.particle.FickianDiffusion(
                        self.param, domain, self.options, phase=phase, x_average=True
                    )
                elif particle in [
                    "uniform profile",
                    "quadratic profile",
                    "quartic profile",
                ]:
                    submod = pybamm.particle.XAveragedPolynomialProfile(
                        self.param, domain, self.options, phase=phase
                    )
                elif particle == "MSMR":
                    submod = pybamm.particle.MSMRDiffusion(
                        self.param, domain, self.options, phase=phase, x_average=True
                    )
                    # also set the submodel for calculating stoichiometry from
                    # potential
                    self.submodels[f"{domain} {phase} stoichiometry"] = (
                        pybamm.particle.MSMRStoichiometryVariables(
                            self.param,
                            domain,
                            self.options,
                            phase=phase,
                            x_average=True,
                        )
                    )
                self.submodels[f"{domain} {phase} particle"] = submod
                self.submodels[f"{domain} {phase} total particle concentration"] = (
                    pybamm.particle.TotalConcentration(
                        self.param, domain, self.options, phase
                    )
                )

    def set_electrolyte_concentration_submodel(self):
        self.submodels["electrolyte diffusion"] = (
            pybamm.electrolyte_diffusion.ConstantConcentration(self.param)
        )



#
# Class to calculate total particle concentrations
#
import pybamm

from .base_particle import BaseParticle


class TotalConcentration(BaseParticle):
    """
    Class to calculate total particle concentrations

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain of the model either 'Negative' or 'Positive'
    options: dict
        A dictionary of options to be passed to the model.
        See :class:`pybamm.BaseBatteryModel`
    phase : str, optional
        Phase of the particle (default is "primary")
    """

    def get_coupled_variables(self, variables):
        domain, Domain = self.domain_Domain
        phase = self.phase
        phase_name = self.phase_name

        c_s_rav = variables[
            f"R-averaged {domain} {phase_name}particle concentration [mol.m-3]"
        ]
        eps_s = variables[
            f"{Domain} electrode {phase_name}active material volume fraction"
        ]
        eps_s_av = pybamm.x_average(eps_s)
        c_s_vol_av = pybamm.x_average(eps_s * c_s_rav) / eps_s_av
        c_scale = self.phase_param.c_max
        L = self.domain_param.L
        A = self.param.A_cc

        variables.update(
            {
                f"{Domain} electrode {phase_name}stoichiometry": c_s_vol_av / c_scale,
                f"{Domain} electrode {phase_name}volume-averaged "
                "concentration": c_s_vol_av / c_scale,
                f"{Domain} electrode {phase_name}volume-averaged "
                "concentration [mol.m-3]": c_s_vol_av,
                f"Total lithium in {phase} phase in {domain} electrode [mol]"
                "": pybamm.yz_average(c_s_vol_av * eps_s_av) * L * A,
            }
        )
        return variables



#
# Algebraic solver class
#
import casadi
import numpy as np
from scipy import optimize
from scipy.sparse import issparse

import pybamm


class AlgebraicSolver(pybamm.BaseSolver):
    """Solve a discretised model which contains only (time independent) algebraic
    equations using a root finding algorithm.
    Uses scipy.optimize.root.
    Note: this solver could be extended for quasi-static models, or models in
    which the time derivative is manually discretised and results in a (possibly
    nonlinear) algebaric system at each time level.

    Parameters
    ----------
    method : str, optional
        The method to use to solve the system (default is "lm"). If it starts with
        "lsq", least-squares minimization is used. The method for least-squares can be
        specified in the form "lsq_methodname"
    tol : float, optional
        The tolerance for the solver (default is 1e-6).
    extra_options : dict, optional
        Any options to pass to the rootfinder. Vary depending on which method is chosen.
        Please consult `SciPy documentation
        <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.show_options.html>`_
        for details. Addititional options to pass to the solver, by default:

        .. code-block:: python

            extra_options = {
                # Tolerance for termination by the change of the independent variables.
                "xtol": 1e-12,
                # Tolerance for termination by the norm of the gradient.
                "gtol": 1e-12,
            }
    """

    def __init__(self, method="lm", tol=1e-6, extra_options=None):
        super().__init__(method=method)
        self.tol = tol

        default_extra_options = {
            "xtol": 1e-12,
            "gtol": 1e-12,
        }
        extra_options = extra_options or {}
        self.extra_options = default_extra_options | extra_options

        self.name = f"Algebraic solver ({method})"
        self._algebraic_solver = True
        pybamm.citations.register("Virtanen2020")

    def set_up_root_solver(self, model, inputs_dict, t_eval):
        """Create and return a rootfinder object. Not used for `pybamm.AlgebraicSolver`.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate.
        inputs_dict : dict
            Dictionary of inputs.
        t_eval : :class:`numpy.array`, size (k,)
            The times at which to compute the solution.
        """
        pass

    @property
    def tol(self):
        return self._tol

    @tol.setter
    def tol(self, value):
        self._tol = value

    def _integrate(self, model, t_eval, inputs_dict=None, t_interp=None):
        """
        Calculate the solution of the algebraic equations through root-finding

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate.
        t_eval : :class:`numpy.array`, size (k,)
            The times at which to compute the solution
        inputs_dict : dict, optional
            Any input parameters to pass to the model when solving
        """
        inputs_dict = inputs_dict or {}
        if model.convert_to_format == "casadi":
            inputs = casadi.vertcat(*[x for x in inputs_dict.values()])
        else:
            inputs = inputs_dict

        y0 = model.y0
        if isinstance(y0, casadi.DM):
            y0 = y0.full()
        y0 = y0.flatten()

        # The casadi algebraic solver can read rhs equations, but leaves them unchanged
        # i.e. the part of the solution vector that corresponds to the differential
        # equations will be equal to the initial condition provided. This allows this
        # solver to be used for initialising the DAE solvers
        # Split y0 into differential and algebraic
        if model.rhs == {}:
            len_rhs = 0
        else:
            len_rhs = model.rhs_eval(t_eval[0], y0, inputs).shape[0]
        y0_diff, y0_alg = np.split(y0, [len_rhs])

        test_result = model.algebraic_eval(0, y0, inputs)

        if isinstance(test_result, casadi.DM):

            def algebraic(t, y):
                result = model.algebraic_eval(t, y, inputs)
                return result.full().flatten()

        else:

            def algebraic(t, y):
                result = model.algebraic_eval(t, y, inputs)
                return result.flatten()

        y_alg = np.empty((len(y0_alg), len(t_eval)))

        timer = pybamm.Timer()
        integration_time = 0
        for idx, t in enumerate(t_eval):

            def root_fun(y_alg, t=t):
                "Evaluates algebraic using y"
                y = np.concatenate([y0_diff, y_alg])
                out = algebraic(t, y)
                pybamm.logger.debug(
                    f"Evaluating algebraic equations at t={t}, L2-norm is {np.linalg.norm(out)}"
                )
                return out

            jac = model.jac_algebraic_eval
            if jac:
                if issparse(jac(t_eval[0], y0, inputs)):

                    def jac_fn(y_alg, jac=jac):
                        """
                        Evaluates Jacobian using y0_diff (fixed) and y_alg (varying)
                        """
                        y = np.concatenate([y0_diff, y_alg])
                        return jac(0, y, inputs)[:, len_rhs:].toarray()

                else:

                    def jac_fn(y_alg, jac=jac):
                        """
                        Evaluates Jacobian using y0_diff (fixed) and y_alg (varying)
                        """
                        y = np.concatenate([y0_diff, y_alg])
                        return jac(0, y, inputs)[:, len_rhs:]

            else:
                jac_fn = None

            itr = 0
            maxiter = 2
            success = False
            while not success:
                # Methods which use least-squares are specified as either "lsq",
                # which uses the default method, or with "lsq__methodname"
                if self.method.startswith("lsq"):
                    if self.method == "lsq":
                        method = "trf"
                    else:
                        method = self.method[5:]
                    if jac_fn is None:
                        jac_fn = "2-point"
                    timer.reset()
                    sol = optimize.least_squares(
                        root_fun,
                        y0_alg,
                        method=method,
                        ftol=self.tol,
                        jac=jac_fn,
                        bounds=model.bounds,
                        **self.extra_options,
                    )
                    integration_time += timer.time()
                    success |= sol.success
                # Methods which use minimize are specified as either "minimize",
                # which uses the default method, or with "minimize__methodname"
                elif self.method.startswith("minimize"):
                    # Adapt the root function for minimize
                    def root_norm(y):
                        return np.sum(root_fun(y) ** 2)

                    if jac_fn is None:
                        jac_norm = None
                    else:

                        def jac_norm(y, jac_fn=jac_fn):
                            return np.sum(2 * root_fun(y) * jac_fn(y), 0)

                    if self.method == "minimize":
                        method = None
                    else:
                        method = self.method[10:]
                    extra_options = self.extra_options
                    if np.any(model.bounds[0] != -np.inf) or np.any(
                        model.bounds[1] != np.inf
                    ):
                        bounds = [
                            (lb, ub)
                            for lb, ub in zip(
                                model.bounds[0], model.bounds[1], strict=True
                            )
                        ]
                    else:
                        bounds = None

                    hess = extra_options.get("hess", None)
                    hessp = extra_options.get("hessp", None)
                    constraints = extra_options.get("constraints", ())
                    callback = extra_options.get("callback", None)

                    timer.reset()
                    sol = optimize.minimize(
                        root_norm,
                        y0_alg,
                        method=method,
                        tol=self.tol,
                        jac=jac_norm,
                        bounds=bounds,
                        hess=hess,
                        hessp=hessp,
                        constraints=constraints,
                        callback=callback,
                        options=extra_options,
                    )
                    integration_time += timer.time()
                    # The solution.success flag is unreliable, so we ignore
                    # it and use the norm of the residual instead
                else:
                    timer.reset()
                    sol = optimize.root(
                        root_fun,
                        y0_alg,
                        method=self.method,
                        tol=self.tol,
                        jac=jac_fn,
                        options=self.extra_options,
                    )
                    integration_time += timer.time()
                    # The solution.success flag is unreliable, so we ignore
                    # it and use the norm of the residual instead

                y0_alg = sol.x
                success |= np.all(abs(sol.fun) < self.tol)
                if success:
                    y_alg[:, idx] = y0_alg
                    break

                if itr > maxiter:
                    raise pybamm.SolverError(
                        "Could not find acceptable solution: solver terminated "
                        "unsuccessfully and maximum solution error "
                        f"({np.max(abs(sol.fun))}) above tolerance ({self.tol})"
                    )
                itr += 1

        # Concatenate differential part
        y_diff = np.r_[[y0_diff] * len(t_eval)].T
        y_sol = np.r_[y_diff, y_alg]
        # Return solution object (no events, so pass None to t_event, y_event)
        sol = pybamm.Solution(
            t_eval, y_sol, model, inputs_dict, termination="final time"
        )
        sol.integration_time = integration_time
        return sol



from .base_porosity import BaseModel
from .constant_porosity import Constant
from .reaction_driven_porosity import ReactionDriven
from .reaction_driven_porosity_ode import ReactionDrivenODE

__all__ = ['base_porosity', 'constant_porosity', 'reaction_driven_porosity',
           'reaction_driven_porosity_ode']



#
# Class for converting SEI concentration into thickness
#
import pybamm

from .base_sei import BaseModel


class SEIThickness(BaseModel):
    """
    Class for converting SEI concentration into thickness

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    reaction_loc : str
        Where the reaction happens: "x-average" (SPM, SPMe, etc),
        "full electrode" (full DFN), or "interface" (half-cell model)
    options : dict
        A dictionary of options to be passed to the model.
    phase : str, optional
        Phase of the particle (default is "primary")
    cracks : bool, optional
        Whether this is a submodel for standard SEI or SEI on cracks
    """

    def __init__(self, param, domain, options, phase="primary", cracks=False):
        super().__init__(param, domain, options=options, phase=phase, cracks=cracks)
        if self.options.electrode_types[domain] == "planar":
            self.reaction_loc = "interface"
        elif self.options["x-average side reactions"] == "true":
            self.reaction_loc = "x-average"
        else:
            self.reaction_loc = "full electrode"

    def get_coupled_variables(self, variables):
        """Update variables related to the SEI thickness."""
        domain, Domain = self.domain_Domain
        phase_param = self.phase_param
        reaction_name = self.reaction_name
        SEI_option = getattr(self.options, domain)["SEI"]
        crack_option = getattr(self.options, domain)["SEI on cracks"]
        if self.options["working electrode"] != "both" and domain == "negative":
            crack_option = "false"  # required if SEI on cracks is used for half-cells

        # Set scales to one for the "no SEI" model so that they are not required
        # by parameter values in general
        if SEI_option == "none":
            c_to_L = 1
            R_sei = 1
        else:
            if self.reaction_loc == "interface":
                c_to_L = phase_param.V_bar_sei
            else:
                a = variables[
                    f"{Domain} electrode {self.phase_name}"
                    "surface area to volume ratio [m-1]"
                ]
                c_to_L = phase_param.V_bar_sei / a
            R_sei = phase_param.R_sei

        if self.reaction_loc == "interface":
            # c_sei is an interfacial quantity [mol.m-2]
            c_sei = variables[f"{Domain} {reaction_name}concentration [mol.m-2]"]
        else:
            # c_sei is a bulk quantity [mol.m-3]
            c_sei = variables[f"{Domain} {reaction_name}concentration [mol.m-3]"]

        if self.reaction == "SEI":
            L_sei = c_sei * c_to_L  # SEI thickness
            if self.size_distribution:
                L_sei_sav = pybamm.size_average(L_sei)  # size-averaged SEI thickness
                L_sei_xav = pybamm.x_average(L_sei_sav)  # x-averaged SEI thickness
            else:
                L_sei_xav = pybamm.x_average(L_sei)  # x-averaged SEI thickness
            L_sei_av = pybamm.yz_average(L_sei_xav)  # volume-averaged SEI thickness

            variables.update(
                {
                    f"X-averaged {self.domain} electrode resistance [Ohm.m2]": L_sei_xav
                    * R_sei,
                }
            )
        # Thickness variables are handled slightly differently for SEI on cracks
        elif self.reaction == "SEI on cracks":
            # if SEI on cracks is false, skip over roughness to avoid division by zero
            if crack_option == "false":
                L_sei = c_sei * c_to_L
            else:
                roughness = variables[f"{Domain} electrode roughness ratio"]
                L_sei = c_sei * c_to_L / (roughness - 1)  # SEI on cracks thickness
            if self.size_distribution:
                L_sei_sav = pybamm.size_average(L_sei)
                L_sei_xav = pybamm.x_average(L_sei_sav)
            else:
                L_sei_xav = pybamm.x_average(L_sei)  # x-average SEI on cracks thickness
            L_sei_av = pybamm.yz_average(L_sei_xav)  # average SEI on cracks thickness

        variables.update(
            {
                f"{Domain} {reaction_name}thickness [m]": L_sei,
                f"X-averaged {domain} {reaction_name}thickness [m]": L_sei_xav,
                f"Volume-averaged {domain} {reaction_name}thickness [m]": L_sei_av,
            }
        )

        return variables



#
# Standard constants
#
from scipy import constants

from pybamm import Constant

R = Constant(constants.R, "R")
F = Constant(constants.physical_constants["Faraday constant"][0], "F")
k_b = Constant(constants.physical_constants["Boltzmann constant"][0], "k_b")
q_e = Constant(constants.physical_constants["electron volt"][0], "q_e")



from collections import OrderedDict


class LRUDict(OrderedDict):
    """LRU extension of a dictionary"""

    def __init__(self, maxsize=None):
        """maxsize limits the item count based on an LRU strategy

        The dictionary remains unbound when maxsize = 0 | None
        """
        super().__init__()
        self.maxsize = maxsize

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        while self.maxsize and self.__len__() > self.maxsize:
            self.popitem(last=False)

    def __getitem__(self, key):
        try:
            self.move_to_end(key, last=True)
        except KeyError:
            pass  # Allow parent to handle exception
        return super().__getitem__(key)

    def get(self, key):
        try:
            self.move_to_end(key, last=True)
        except KeyError:
            pass  # Allow parent to handle exception
        return super().get(key)



import os

import numpy as np

import pybamm
from pybamm.util import import_optional_dependency


class ScikitFemGenerator3D(pybamm.MeshGenerator):
    """
    A mesh generator that creates 3D tetrahedral meshes using scikit-fem.

    Parameters
    ----------
    geom_type : str
        The type of geometry to generate. Must be one of "pouch" for a rectangular
        prism, or "cylinder" for a cylindrical annulus.
    gen_params : dict
        A dictionary of geometry-specific parameters. for eg:

        - "h" : float, optional
            The target characteristic length of the mesh elements. A smaller 'h'
            results in a finer mesh and a more accurate solution, but
            increases computation time. Default is 0.3.
    """

    def __init__(self, geom_type, **gen_params):
        supported_geometries = ["pouch", "cylinder"]
        if geom_type not in supported_geometries:
            raise pybamm.GeometryError(
                f"geom_type must be one of {supported_geometries}, not '{geom_type}'"
            )
        super().__init__(ScikitFemSubMesh3D)
        self.geom_type = geom_type
        self.gen_params = gen_params

    def _make_pouch_mesh(self, x_lim, y_lim, z_lim, h):
        """
        Create a structured pouch mesh.

        Parameters
        ----------
        x_lim, y_lim, z_lim : tuple
            Domain limits for each dimension
        h : float
            Target mesh size

        Returns
        -------
        skfem.MeshTet
            Pouch mesh with proper boundary tags
        """
        skfem = import_optional_dependency("skfem")
        x_min, x_max = x_lim
        y_min, y_max = y_lim
        z_min, z_max = z_lim

        nx = max(2, int(np.ceil((x_max - x_min) / h)) + 1)
        ny = max(2, int(np.ceil((y_max - y_min) / h)) + 1)
        nz = max(2, int(np.ceil((z_max - z_min) / h)) + 1)

        mesh = skfem.MeshTet.init_tensor(
            np.linspace(x_min, x_max, nx),
            np.linspace(y_min, y_max, ny),
            np.linspace(z_min, z_max, nz),
        )

        bnd_facets = mesh.boundary_facets()
        midpoints = mesh.p[:, mesh.facets[:, bnd_facets]].mean(axis=1)

        boundaries = {
            "x_min": bnd_facets[np.isclose(midpoints[0], x_min)],
            "x_max": bnd_facets[np.isclose(midpoints[0], x_max)],
            "y_min": bnd_facets[np.isclose(midpoints[1], y_min)],
            "y_max": bnd_facets[np.isclose(midpoints[1], y_max)],
            "z_min": bnd_facets[np.isclose(midpoints[2], z_min)],
            "z_max": bnd_facets[np.isclose(midpoints[2], z_max)],
        }

        subdomains = {"default": np.arange(mesh.nelements)}
        return mesh.with_boundaries(boundaries).with_subdomains(subdomains)

    def _make_cylindrical_mesh(self, r_lim, z_lim, h):
        """
        Create a cylindrical annulus mesh.

        Parameters
        ----------
        r_lim : tuple
            Radial limits (inner radius, outer radius)
        z_lim : tuple
            Vertical limits (z_min, z_max)
        h : float
            Target mesh size

        Returns
        -------
        skfem.MeshTet
            Cylindrical mesh with proper boundary tags
        """
        skfem = import_optional_dependency("skfem")
        from scipy.spatial import Delaunay

        r_inner, r_outer = r_lim
        z_min, z_max = z_lim
        height = z_max - z_min

        n_radial = max(4, int((r_outer - r_inner) / h) + 1)
        n_theta_base = max(16, int(2 * np.pi * r_outer / h))
        r_coords = np.linspace(r_inner, r_outer, n_radial)
        points_list = []
        for r in r_coords:
            if r > 1e-9:
                n_theta = max(8, int(n_theta_base * (r / r_outer)))
                thetas = np.linspace(0, 2 * np.pi, n_theta, endpoint=False)
                for theta in thetas:
                    points_list.append([r * np.cos(theta), r * np.sin(theta)])
        points_2d = np.array(points_list)
        tri_2d = Delaunay(points_2d)
        triangles_base = tri_2d.simplices

        n_z = max(5, int(height / h))
        z_coords = np.linspace(z_min, z_max, n_z)
        n_nodes_per_layer = points_2d.shape[0]
        nodes_3d = np.zeros((n_nodes_per_layer * n_z, 3))
        for i, z in enumerate(z_coords):
            start, end = i * n_nodes_per_layer, (i + 1) * n_nodes_per_layer
            nodes_3d[start:end, :2] = points_2d
            nodes_3d[start:end, 2] = z
        tetrahedra = []
        for i in range(n_z - 1):
            for tri in triangles_base:
                v = [node_idx + (i * n_nodes_per_layer) for node_idx in tri]
                v_top = [node_idx + ((i + 1) * n_nodes_per_layer) for node_idx in tri]
                tetrahedra.extend(
                    [
                        [v[0], v[1], v[2], v_top[2]],
                        [v[0], v[1], v_top[1], v_top[2]],
                        [v[0], v_top[0], v_top[1], v_top[2]],
                    ]
                )
        mesh = skfem.MeshTet(nodes_3d.T, np.array(tetrahedra).T)

        bottom_nodes = set(range(n_nodes_per_layer))
        top_nodes = set(range(n_nodes_per_layer * (n_z - 1), n_nodes_per_layer * n_z))

        points_2d_r = np.linalg.norm(points_2d, axis=1)
        tol = h * 0.1
        inner_nodes_2d = np.where(np.isclose(points_2d_r, r_inner, atol=tol))[0]
        outer_nodes_2d = np.where(np.isclose(points_2d_r, r_outer, atol=tol))[0]

        inner_wall_nodes = {
            n + (i * n_nodes_per_layer) for i in range(n_z) for n in inner_nodes_2d
        }
        outer_wall_nodes = {
            n + (i * n_nodes_per_layer) for i in range(n_z) for n in outer_nodes_2d
        }

        bottom_facets, top_facets, inner_facets, outer_facets = [], [], [], []

        for facet_idx in mesh.boundary_facets():
            facet_nodes = set(mesh.facets[:, facet_idx])
            if facet_nodes.issubset(bottom_nodes):
                bottom_facets.append(facet_idx)
            elif facet_nodes.issubset(top_nodes):
                top_facets.append(facet_idx)
            elif facet_nodes.issubset(inner_wall_nodes):
                inner_facets.append(facet_idx)
            elif facet_nodes.issubset(outer_wall_nodes):
                outer_facets.append(facet_idx)

        boundaries = {}
        if len(bottom_facets) > 0:
            boundaries["z_min"] = np.array(bottom_facets)
        if len(top_facets) > 0:
            boundaries["z_max"] = np.array(top_facets)
        if len(inner_facets) > 0:
            boundaries["r_min"] = np.array(inner_facets)
        if len(outer_facets) > 0:
            boundaries["r_max"] = np.array(outer_facets)

        return mesh.with_boundaries(boundaries)

    def __call__(self, lims, npts):
        """
        Main entry point called by PyBaMM's Discretisation.

        Parameters
        ----------
        lims : dict
            Dictionary of domain limits
        npts : dict
            Dictionary of target number of points

        Returns
        -------
        ScikitFemSubMesh3D
            Generated 3D submesh
        """
        h = float(self.gen_params.get("h", 0.3))

        lims_dict = {}
        for var, lim in lims.items():
            if isinstance(var, pybamm.SpatialVariable):
                lims_dict[var.name] = (lim["min"], lim["max"])
            else:
                lims_dict[var] = (float(lim["min"]), float(lim["max"]))

        if self.geom_type == "pouch":
            x_lim = lims_dict["x"]
            y_lim = lims_dict["y"]
            z_lim = lims_dict["z"]
            coord_sys = "cartesian"
            mesh = self._make_pouch_mesh(x_lim, y_lim, z_lim, h)

        elif self.geom_type == "cylinder":
            if "r" in lims_dict:
                r_lim = lims_dict["r"]
            else:
                r_lim = lims_dict["r_macro"]

            z_lim = lims_dict["z"]
            coord_sys = "cylindrical polar"
            mesh = self._make_cylindrical_mesh(r_lim, z_lim, h)
        else:
            raise ValueError(f"Unknown geom_type: {self.geom_type}")  # pragma: no cover

        if mesh is None:
            raise pybamm.DiscretisationError(
                f"Mesh generation failed for {self.geom_type}"
            )  # pragma: no cover

        nodes = mesh.p.T
        elements = mesh.t.T
        submesh = self.submesh_type(mesh, nodes, elements, coord_sys)
        return submesh


class UserSuppliedSubmesh3D(pybamm.MeshGenerator):
    """
    A mesh generator that loads a 3D mesh from an external file (e.g., VTK).

    This allows using high-quality meshes from any external tool. The mesh
    file must contain integer tags identifying the different physical domains
    and boundaries. This generator works directly with your existing
    'ScikitFemSubMesh3D' and 'ScikitFiniteElement3D' classes.

    Parameters
    ----------
    file_path : str
        Path to the mesh file.
    boundary_mapping : dict
        Maps PyBaMM boundary names to the integer tags in the mesh file.
    domain_mapping : dict
        Maps PyBaMM domain names to the integer tags in the mesh file.
    domain_tag_name : str, optional
        The name of the data array in the mesh file for domain tags.
    boundary_tag_name : str, optional
        The name of the data array in the mesh file for boundary tags.
    coord_sys : str, optional
        The coordinate system ("cartesian" or "cylindrical polar").

    Notes
    -----
    The external mesh file must meet the following criteria to be compatible:

        1.  **Element Types**: The 3D volume must be meshed with 4-node
            **tetrahedral** elements. The 2D boundaries must be meshed with
            3-node **triangular** elements.

        2.  **Physical Tags**: The file must contain integer tags to identify
            physical groups.
            - Each tetrahedron in the volume must be tagged with an integer
            that corresponds to a PyBaMM domain.
            - Each triangle on a boundary must be tagged with an integer that
            corresponds to a PyBaMM boundary.

        3.  **File Format**: Any format supported by the `meshio` library is
            acceptable. For maximum reliability in preserving physical tags,
            the **Gmsh `.msh` Version 2.2 (ASCII)** format is recommended.
    """

    def __init__(
        self,
        file_path,
        boundary_mapping,
        domain_mapping,
        domain_tag_name=None,
        boundary_tag_name=None,
        coord_sys=None,
    ):
        super().__init__(ScikitFemSubMesh3D)
        self.file_path = file_path
        self.boundary_mapping = boundary_mapping
        self.domain_mapping = domain_mapping
        self.domain_tag_name = domain_tag_name
        self.boundary_tag_name = boundary_tag_name

        if coord_sys:
            self.coord_sys = coord_sys
        else:
            if any("r_min" in k or "r_max" in k for k in boundary_mapping.keys()):
                self.coord_sys = "cylindrical polar"
            else:
                self.coord_sys = "cartesian"

    def __call__(self, lims, npts):
        skfem_mesh = ScikitFemSubMesh3D.load_mesh_from_file(
            self.file_path,
            self.boundary_mapping,
            self.domain_mapping,
            self.domain_tag_name,
            self.boundary_tag_name,
        )
        nodes = skfem_mesh.p.T
        elements = skfem_mesh.t.T
        return self.submesh_type(skfem_mesh, nodes, elements, self.coord_sys)


class ScikitFemSubMesh3D(pybamm.SubMesh):
    """
    A 3D submesh class for unstructured tetrahedral meshes generated by scikit-fem.

    Parameters
    ----------
    skfem_mesh : skfem.MeshTet
        The scikit-fem mesh object
    nodes : array_like
        Array of node coordinates (npts, 3)
    elements : array_like
        Array of element connectivity (nelements, 4)
    coord_sys : str
        The coordinate system of the submesh
    tabs : dict, optional
        Information about tabs (unused in 3D)
    """

    def __init__(self, skfem_mesh, nodes, elements, coord_sys, tabs=None):
        super().__init__()
        skfem = import_optional_dependency("skfem")

        self._skfem_mesh = skfem_mesh
        self.nodes = nodes
        self.elements = elements
        self.coord_sys = coord_sys
        self.tabs = tabs
        self.dimension = 3
        self.npts = nodes.shape[0]
        self.nelements = elements.shape[0]
        self.edges = self._skfem_mesh.edges.T

        self.basis = skfem.InteriorBasis(self._skfem_mesh, skfem.ElementTetP1())
        self.facet_basis = skfem.FacetBasis(self._skfem_mesh, self.basis.elem)
        self.element = skfem.ElementTetP1()

        if (
            hasattr(self._skfem_mesh, "boundaries")
            and self._skfem_mesh.boundaries is not None
        ):
            for name in self._skfem_mesh.boundaries:
                facet_basis = skfem.FacetBasis(
                    self._skfem_mesh,
                    self.basis.elem,
                    facets=self._skfem_mesh.boundaries[name],
                )
                setattr(self, f"{name}_basis", facet_basis)

                dofs = self.basis.get_dofs(name).all()
                setattr(self, f"{name}_dofs", dofs)

                normals = facet_basis.normals
                setattr(self, f"{name}_normals", normals)

        self._compute_element_volumes()
        self._compute_element_centroids()
        self.internal_boundaries = []

    def _compute_element_volumes(self):
        """Compute volumes of all tetrahedral elements."""
        vertices = self.nodes[self.elements]
        v1 = vertices[:, 1] - vertices[:, 0]
        v2 = vertices[:, 2] - vertices[:, 0]
        v3 = vertices[:, 3] - vertices[:, 0]
        self.element_volumes = np.abs(np.einsum("ij,ij->i", np.cross(v1, v2), v3)) / 6

    def _compute_element_centroids(self):
        """Compute centroids of all tetrahedral elements."""
        self.element_centroids = np.mean(self.nodes[self.elements], axis=1)

    def to_json(self):
        """Convert mesh to JSON format."""
        return {
            "mesh_type": self.__class__.__name__,
            "nodes": self.nodes.tolist(),
            "elements": self.elements.tolist(),
            "coord_sys": self.coord_sys,
            "element_volumes": self.element_volumes.tolist(),
            "element_centroids": self.element_centroids.tolist(),
        }

    @classmethod
    def _from_json(cls, json_dict):
        """Create mesh from JSON format."""
        skfem = import_optional_dependency("skfem")
        nodes = np.array(json_dict["nodes"])
        elements = np.array(json_dict["elements"])
        mesh = skfem.MeshTet(nodes.T, elements.T)
        return cls(mesh, nodes, elements, json_dict["coord_sys"])

    @staticmethod
    def load_mesh_from_file(
        file_path,
        boundary_mapping,
        domain_mapping,
        domain_tag_name=None,
        boundary_tag_name=None,
    ):
        """
        Loads a generic mesh file and converts it to a scikit-fem Mesh object.

        This function is designed to work with meshes from most software packages
        including GMSH, Ansys, Abaqus, FEniCS, etc. It automatically detects
        common tag naming conventions or allows manual specification.

        Parameters
        ----------
        file_path : str
            The path to the mesh file (It supports .msh, .xdmf, may also work with others but not tested).
        boundary_mapping : dict
            Maps PyBaMM boundary names to integer tags (e.g., {"r_inner": 1}).
        domain_mapping : dict
            Maps PyBaMM domain names to integer tags (e.g., {"current collector": 5}).
        domain_tag_name : str, optional
            The name of the cell data array that contains domain tags for elements.
        boundary_tag_name : str, optional
            The name of the cell data array that contains boundary tags for facets.
        """
        meshio = pybamm.util.import_optional_dependency("meshio")
        skfem = pybamm.util.import_optional_dependency("skfem")

        tested_formats = [".msh", ".xdmf"]
        _, file_ext = os.path.splitext(file_path)
        if file_ext not in tested_formats:  # pragma: no cover
            pybamm.logger.warning(
                f"File format '{file_ext}' has not been explicitly tested and may not "
                "work correctly. Recommended formats are .msh and .xdmf."
            )

        try:
            m = meshio.read(file_path)
        except Exception as e:
            raise pybamm.GeometryError(
                f"Could not read mesh file '{file_path}': {e}"
            ) from e

        nodes = m.points
        if nodes.shape[1] != 3:
            raise pybamm.GeometryError(
                f"Mesh must be 3D, got points with shape {nodes.shape}"
            )

        def find_tag_name(
            user_provided_name, default_candidates, available_data, context=""
        ):
            """
            Enhanced tag name finder that works with most mesh formats.
            """
            # If user explicitly provided a name, use it
            if user_provided_name:
                if user_provided_name in available_data:
                    return user_provided_name
                else:
                    available_keys = list(available_data.keys())
                    raise pybamm.GeometryError(
                        f"User-specified {context} tag name '{user_provided_name}' not found. "
                        f"Available arrays: {available_keys}"
                    )

            for candidate in default_candidates:
                if candidate in available_data:
                    return candidate

            integer_arrays = [
                name
                for name, data in available_data.items()
                if np.issubdtype(data.dtype, np.integer)
            ]

            if len(integer_arrays) == 1:
                return integer_arrays[0]

            priority_patterns = ["physical", "tag", "id", "group", "material", "region"]
            for pattern in priority_patterns:
                for name in integer_arrays:
                    if pattern.lower() in name.lower():
                        return name

            raise pybamm.GeometryError(
                f"Could not automatically detect {context} tag array in '{file_path}'. "
                f"Available arrays: {list(available_data.keys())}. "
                f"Integer arrays found: {integer_arrays}. "
                f"Please specify manually using '{context.replace(' ', '_')}_tag_name'."
            )

        try:
            tet_cells = m.get_cells_type("tetra")
            if len(tet_cells) == 0:
                raise pybamm.GeometryError(
                    f"No tetrahedral elements found in '{file_path}'"
                )

            tet_cell_index = None
            for i, cell_block in enumerate(m.cells):
                if cell_block.type == "tetra":
                    tet_cell_index = i
                    break

            if tet_cell_index is None:
                raise pybamm.GeometryError("Could not find tetrahedra in cell list")

            tet_cell_data_dict = {}
            for name, data_list in m.cell_data.items():
                if len(data_list) > tet_cell_index:
                    data = data_list[tet_cell_index]
                    if len(data) == len(tet_cells):
                        tet_cell_data_dict[name] = data

            domain_tag_name = find_tag_name(
                domain_tag_name,
                [
                    "gmsh:physical",
                    "physical",
                    "PhysicalGroup",
                    "MaterialId",
                    "CellEntityIds",
                    "gmsh:geometrical",
                ],
                tet_cell_data_dict,
                "domain",
            )

            elements = tet_cells
            tet_cell_data = tet_cell_data_dict[domain_tag_name]

            subdomains = {}
            for name, tag in domain_mapping.items():
                matching_elements = np.where(tet_cell_data == tag)[0]
                if len(matching_elements) == 0:
                    pybamm.logger.warning(
                        f"No elements found for domain '{name}' with tag {tag}"
                    )
                else:
                    subdomains[name] = matching_elements
                    pybamm.logger.info(
                        f"Found {len(matching_elements)} elements for domain '{name}' (tag {tag})"
                    )

        except Exception as e:
            raise pybamm.GeometryError(
                f"Failed to extract tetrahedral elements from '{file_path}': {e}"
            ) from e

        skfem_mesh = skfem.MeshTet(nodes.T, elements.T)

        boundaries = {}
        try:
            tri_cells = m.get_cells_type("triangle")

            if len(tri_cells) > 0:
                tri_cell_indices = []
                for i, cell_block in enumerate(m.cells):
                    if cell_block.type == "triangle":
                        tri_cell_indices.append(i)

                # Combine all triangle cell data
                all_tri_cells = []
                all_tri_data = []

                for idx in tri_cell_indices:
                    block_cells = m.cells[idx].data
                    all_tri_cells.extend(block_cells)

                    for name, data_list in m.cell_data.items():
                        if len(data_list) > idx:
                            if name not in [arr_name for arr_name, _ in all_tri_data]:
                                all_tri_data.append((name, []))
                            for arr_name, arr_data in all_tri_data:
                                if arr_name == name:
                                    arr_data.extend(data_list[idx])
                                    break

                all_tri_cells = np.array(all_tri_cells)

                tri_cell_data_dict = {}
                for name, data in all_tri_data:
                    tri_cell_data_dict[name] = np.array(data)

                if tri_cell_data_dict:
                    # Find boundary tag name
                    boundary_tag_name = find_tag_name(
                        boundary_tag_name,
                        [
                            "gmsh:physical",
                            "physical",
                            "PhysicalGroup",
                            "FaceEntityIds",
                            "gmsh:geometrical",
                        ],
                        tri_cell_data_dict,
                        "boundary",
                    )

                    tri_cell_data = tri_cell_data_dict[boundary_tag_name]

                    # Map triangle facets to mesh boundary facets
                    bnd_facets_indices = skfem_mesh.boundary_facets()
                    bnd_facets_nodes = skfem_mesh.facets[:, bnd_facets_indices]

                    # Create mapping from node sets to facet indices
                    facet_map = {}
                    for i in range(len(bnd_facets_indices)):
                        node_set = frozenset(bnd_facets_nodes[:, i])
                        facet_map[node_set] = bnd_facets_indices[i]

                    # Find boundary facets for each tag
                    for name, tag in boundary_mapping.items():
                        tagged_triangles = all_tri_cells[tri_cell_data == tag]
                        facet_indices = []

                        for tri in tagged_triangles:
                            tri_set = frozenset(tri)
                            if tri_set in facet_map:
                                facet_indices.append(facet_map[tri_set])

                        if facet_indices:
                            boundaries[name] = np.array(facet_indices)
                            pybamm.logger.info(
                                f"Found {len(facet_indices)} boundary facets for '{name}' (tag {tag})"
                            )
                        else:
                            pybamm.logger.warning(
                                f"No boundary facets found for '{name}' with tag {tag}"
                            )

        except Exception as e:
            pybamm.logger.warning(
                f"Could not extract boundary information from '{file_path}': {e}"
            )

        if boundaries:
            skfem_mesh = skfem_mesh.with_boundaries(boundaries)
        if subdomains:
            skfem_mesh = skfem_mesh.with_subdomains(subdomains)

        pybamm.logger.info(
            f"Successfully loaded mesh: {len(nodes)} nodes, {len(elements)} elements, "
            f"{len(boundaries)} boundary groups, {len(subdomains)} subdomains"
        )

        return skfem_mesh



#
# Concatenation classes
#
from __future__ import annotations

import copy
from collections import defaultdict
from collections.abc import Sequence
from typing import Any

import numpy as np
import numpy.typing as npt
import sympy
from scipy.sparse import issparse, vstack

import pybamm


class Concatenation(pybamm.Symbol):
    """
    A node in the expression tree representing a concatenation of symbols.

    Parameters
    ----------
    children : iterable of :class:`pybamm.Symbol`
        The symbols to concatenate
    """

    def __init__(
        self,
        *children: pybamm.Symbol,
        name: str | None = None,
        check_domain=True,
        concat_fun=None,
    ):
        # The second condition checks whether this is the base Concatenation class
        # or a subclass of Concatenation
        # (ConcatenationVariable, NumpyConcatenation, ...)
        if all(isinstance(child, pybamm.Variable) for child in children) and issubclass(
            Concatenation, type(self)
        ):
            raise TypeError(
                "'ConcatenationVariable' should be used for concatenating 'Variable' "
                "objects. We recommend using the 'concatenation' function, which will "
                "automatically choose the best form."
            )
        if name is None:
            name = "concatenation"
        if check_domain:
            domains = self.get_children_domains(children)
        else:
            domains = {"primary": []}
        self.concatenation_function = concat_fun

        super().__init__(name, children, domains=domains)

    @classmethod
    def _from_json(cls, snippet: dict):
        """Creates a new Concatenation instance from a json object"""
        instance = cls.__new__(cls)

        instance.concatenation_function = snippet["concat_fun"]

        super(Concatenation, instance).__init__(
            snippet["name"], tuple(snippet["children"]), domains=snippet["domains"]
        )

        return instance

    def __str__(self):
        """See :meth:`pybamm.Symbol.__str__()`."""
        out = self.name + "("
        for child in self.children:
            out += f"{child!s}, "
        out = out[:-2] + ")"
        return out

    def _diff(self, variable: pybamm.Symbol):
        """See :meth:`pybamm.Symbol._diff()`."""
        children_diffs = [child.diff(variable) for child in self.children]
        if len(children_diffs) == 1:
            diff = children_diffs[0]
        else:
            diff = self.__class__(*children_diffs)

        return diff

    def get_children_domains(self, children: Sequence[pybamm.Symbol]):
        # combine domains from children
        domain: list = []
        for child in children:
            if not isinstance(child, pybamm.Symbol):
                raise TypeError(f"{child} is not a pybamm symbol")
            child_domain = child.domain
            if child_domain == []:
                raise pybamm.DomainError(
                    f"Cannot concatenate child '{child}' with empty domain"
                )
            if set(domain).isdisjoint(child_domain):
                domain += child_domain
            else:
                raise pybamm.DomainError("domain of children must be disjoint")

        auxiliary_domains = children[0].domains
        for level, dom in auxiliary_domains.items():
            if level != "primary" and dom != []:
                for child in children[1:]:
                    if child.domains[level] not in [dom, []]:
                        raise pybamm.DomainError(
                            "children must have same or empty auxiliary domains"
                        )

        domains = {**auxiliary_domains, "primary": domain}

        return domains

    def _concatenation_evaluate(self, children_eval: list[npt.NDArray[Any]]):
        """See :meth:`Concatenation._concatenation_evaluate()`."""
        if len(children_eval) == 0:
            return np.array([])
        else:
            return self.concatenation_function(children_eval)

    def evaluate(
        self,
        t: float | None = None,
        y: npt.NDArray[np.float64] | None = None,
        y_dot: npt.NDArray[np.float64] | None = None,
        inputs: dict | str | None = None,
    ):
        """See :meth:`pybamm.Symbol.evaluate()`."""
        children_eval = [child.evaluate(t, y, y_dot, inputs) for child in self.children]
        return self._concatenation_evaluate(children_eval)

    def create_copy(
        self,
        new_children: list[pybamm.Symbol] | None = None,
        perform_simplifications: bool = True,
    ):
        """See :meth:`pybamm.Symbol.new_copy()`."""
        children = self._children_for_copying(new_children)

        return self._concatenation_new_copy(children, perform_simplifications)

    def _concatenation_new_copy(self, children, perform_simplifications: bool = True):
        """
        Creates a copy for the current concatenation class using the convenience
        function :meth:`concatenation` to perform simplifications based on the new
        children before creating the new copy.
        """
        if perform_simplifications:
            return concatenation(*children, name=self.name)
        else:
            return self.__class__(*children, name=self.name)

    def _concatenation_jac(self, children_jacs):
        """Calculate the Jacobian of a concatenation."""
        raise NotImplementedError

    def _evaluate_for_shape(self):
        """See :meth:`pybamm.Symbol.evaluate_for_shape`"""
        if len(self.children) == 0:
            return np.array([])
        else:
            # Default: use np.concatenate
            concatenation_function = self.concatenation_function or np.concatenate
            return concatenation_function(
                [child.evaluate_for_shape() for child in self.children]
            )

    def is_constant(self):
        """See :meth:`pybamm.Symbol.is_constant()`."""
        return all(child.is_constant() for child in self.children)

    def _sympy_operator(self, *children):
        """Apply appropriate SymPy operators."""
        self.concat_latex = tuple(map(sympy.latex, children))

        if self.print_name is not None:
            return sympy.Symbol(self.print_name)
        else:
            concat_str = r"\\".join(self.concat_latex)
            concat_sym = sympy.Symbol(r"\begin{cases}" + concat_str + r"\end{cases}")
            return concat_sym

    def to_equation(self):
        """Convert the node and its subtree into a SymPy equation."""
        eq_list = []
        for child in self.children:
            eq = child.to_equation()
            eq_list.append(eq)
        return self._sympy_operator(*eq_list)


class NumpyConcatenation(Concatenation):
    """
    A node in the expression tree representing a concatenation of equations, when we
    *don't* care about domains. The class :class:`pybamm.DomainConcatenation`, which
    *is* careful about domains and uses broadcasting where appropriate, should be used
    whenever possible instead.

    Upon evaluation, equations are concatenated using numpy concatenation.

    Parameters
    ----------
    children : iterable of :class:`pybamm.Symbol`
        The equations to concatenate
    """

    def __init__(self, *children: pybamm.Symbol):
        children = list(children)
        # Turn objects that evaluate to scalars to objects that evaluate to vectors,
        # so that we can concatenate them
        for i, child in enumerate(children):
            if child.evaluates_to_number():
                children[i] = child * pybamm.Vector([1])
        super().__init__(
            *children,
            name="numpy_concatenation",
            check_domain=False,
            concat_fun=np.concatenate,
        )

    @classmethod
    def _from_json(cls, snippet: dict):
        """See :meth:`pybamm.Concatenation._from_json()`."""

        snippet["name"] = "numpy_concatenation"
        snippet["concat_fun"] = np.concatenate

        instance = super()._from_json(snippet)

        return instance

    def _concatenation_jac(self, children_jacs):
        """See :meth:`pybamm.Concatenation.concatenation_jac()`."""
        children = self.children
        if len(children) == 0:
            return pybamm.Scalar(0)
        else:
            return SparseStack(*children_jacs)

    def _concatenation_new_copy(
        self,
        children,
        perform_simplifications: bool = True,
    ):
        """See :meth:`pybamm.Concatenation._concatenation_new_copy()`."""
        if perform_simplifications:
            return numpy_concatenation(*children)
        else:
            raise NotImplementedError(
                f"{self.__class__.__name__} should always be copied using "
                "simplification checks"
            )


class DomainConcatenation(Concatenation):
    """
    A node in the expression tree representing a concatenation of symbols, being
    careful about domains.

    It is assumed that each child has a domain, and the final concatenated vector will
    respect the sizes and ordering of domains established in mesh keys

    Parameters
    ----------

    children : iterable of :class:`pybamm.Symbol`
        The symbols to concatenate

    full_mesh : :class:`pybamm.Mesh`
        The underlying mesh for discretisation, used to obtain the number of mesh points
        in each domain.

    copy_this : :class:`pybamm.DomainConcatenation` (optional)
        if provided, this class is initialised by copying everything except the children
        from `copy_this`. `mesh` is not used in this case
    """

    def __init__(
        self,
        children: Sequence[pybamm.Symbol],
        full_mesh: pybamm.Mesh,
        copy_this: pybamm.DomainConcatenation | None = None,
    ):
        # Convert any constant symbols in children to a Vector of the right size for
        # concatenation
        children = list(children)

        # Allow the base class to sort the domains into the correct order
        super().__init__(*children, name="domain_concatenation")

        if copy_this is None:
            # store mesh
            self._full_mesh = full_mesh

            # create dict of domain => slice of final vector
            self.secondary_dimensions_npts = self._get_auxiliary_domain_repeats(
                self.domains
            )
            self._slices = self.create_slices(self)

            # store size of final vector
            self._size = self._slices[self.domain[-1]][-1].stop

            # create disc of domain => slice for each child
            self._children_slices = [
                self.create_slices(child) for child in self.children
            ]
        else:
            self._full_mesh = copy.copy(copy_this._full_mesh)
            self._slices = copy.copy(copy_this._slices)
            self._size = copy.copy(copy_this._size)
            self._children_slices = copy.copy(copy_this._children_slices)
            self.secondary_dimensions_npts = copy_this.secondary_dimensions_npts

    @classmethod
    def _from_json(cls, snippet: dict):
        """See :meth:`pybamm.Concatenation._from_json()`."""

        snippet["name"] = "domain_concatenation"
        snippet["concat_fun"] = None

        instance = super()._from_json(snippet)

        def repack_defaultDict(slices):
            slices = defaultdict(list, slices)
            for domain, sls in slices.items():
                sls = [slice(s["start"], s["stop"], s["step"]) for s in sls]
                slices[domain] = sls
            return slices

        instance._size = snippet["size"]
        instance._slices = repack_defaultDict(snippet["slices"])
        instance._children_slices = [
            repack_defaultDict(s) for s in snippet["children_slices"]
        ]
        instance.secondary_dimensions_npts = snippet["secondary_dimensions_npts"]

        return instance

    def _get_auxiliary_domain_repeats(self, auxiliary_domains: dict) -> int:
        """Helper method to read the 'auxiliary_domain' meshes."""
        mesh_pts = 1
        for level, dom in auxiliary_domains.items():
            if level != "primary" and dom != []:
                mesh_pts *= self.full_mesh[dom].npts
        return mesh_pts

    @property
    def full_mesh(self):
        return self._full_mesh

    def create_slices(self, node: pybamm.Symbol) -> defaultdict:
        slices = defaultdict(list)
        start = 0
        end = 0
        second_pts = self._get_auxiliary_domain_repeats(self.domains)
        if second_pts != self.secondary_dimensions_npts:
            raise ValueError(
                """Concatenation and children must have the same number of
                points in secondary dimensions"""
            )
        for _ in range(second_pts):
            for dom in node.domain:
                end += self.full_mesh[dom].npts
                slices[dom].append(slice(start, end))
                start = end
        return slices

    def _concatenation_evaluate(self, children_eval: list[npt.NDArray[Any]]):
        """See :meth:`Concatenation._concatenation_evaluate()`."""
        # preallocate vector
        vector = np.empty((self._size, 1))

        # loop through domains of children writing subvectors to final vector
        for child_vector, slices in zip(
            children_eval, self._children_slices, strict=True
        ):
            for child_dom, child_slice in slices.items():
                for i, _slice in enumerate(child_slice):
                    vector[self._slices[child_dom][i]] = child_vector[_slice]

        return vector

    def _concatenation_jac(self, children_jacs):
        """See :meth:`pybamm.Concatenation.concatenation_jac()`."""
        # note that this assumes that the children are in the right order and only have
        # one domain each
        jacs = []
        for i in range(self.secondary_dimensions_npts):
            for child_jac, slices in zip(
                children_jacs, self._children_slices, strict=True
            ):
                if len(slices) > 1:
                    raise NotImplementedError(
                        """jacobian only implemented for when each child has
                        a single domain"""
                    )
                child_slice = next(iter(slices.values()))
                jacs.append(pybamm.Index(child_jac, child_slice[i]))
        return SparseStack(*jacs)

    def _concatenation_new_copy(
        self, children: list[pybamm.Symbol], perform_simplifications: bool = True
    ):
        """See :meth:`pybamm.Concatenation._concatenation_new_copy()`."""
        if perform_simplifications:
            return simplified_domain_concatenation(
                children, self.full_mesh, copy_this=self
            )
        else:
            return DomainConcatenation(children, self.full_mesh, copy_this=self)

    def to_json(self):
        """
        Method to serialise a DomainConcatenation object into JSON.
        """

        def unpack_defaultDict(slices):
            slices = dict(slices)
            for domain, sls in slices.items():
                sls = [{"start": s.start, "stop": s.stop, "step": s.step} for s in sls]
                slices[domain] = sls
            return slices

        json_dict = {
            "name": self.name,
            "id": self.id,
            "domains": self.domains,
            "slices": unpack_defaultDict(self._slices),
            "size": self._size,
            "children_slices": [
                unpack_defaultDict(child_slice) for child_slice in self._children_slices
            ],
            "secondary_dimensions_npts": self.secondary_dimensions_npts,
        }

        return json_dict


class SparseStack(Concatenation):
    """
    A node in the expression tree representing a concatenation of sparse
    matrices. As with NumpyConcatenation, we *don't* care about domains.
    The class :class:`pybamm.DomainConcatenation`, which *is* careful about
    domains and uses broadcasting where appropriate, should be used whenever
    possible instead.

    Parameters
    ----------
    children : iterable of :class:`Concatenation`
        The equations to concatenate
    """

    def __init__(self, *children, name="sparse_stack"):
        children = list(children)
        if not any(issparse(child.evaluate_for_shape()) for child in children):
            concatenation_function = np.vstack
        else:
            concatenation_function = vstack
        super().__init__(
            *children,
            name=name,
            check_domain=False,
            concat_fun=concatenation_function,
        )

    def _concatenation_new_copy(self, children, perform_simplifications=True):
        """See :meth:`pybamm.Concatenation._concatenation_new_copy()`."""
        return SparseStack(*children)


class ConcatenationVariable(Concatenation):
    """A Variable representing a concatenation of variables."""

    def __init__(self, *children, name: str | None = None):
        if name is None:
            # Name is the intersection of the children names (should usually make sense
            # if the children have been named consistently)
            name = intersect(children[0].name, children[1].name)
            for child in children[2:]:
                name = intersect(name, child.name)
            if len(name) == 0:
                name = None
            # name is unchanged if its length is 1
            elif len(name) > 1:
                name = name[0].capitalize() + name[1:]

        if len(children) > 0:
            if all(child.scale == children[0].scale for child in children):
                self._scale = children[0].scale
            else:
                raise ValueError("Cannot concatenate symbols with different scales")
            if all(child.reference == children[0].reference for child in children):
                self._reference = children[0].reference
            else:
                raise ValueError("Cannot concatenate symbols with different references")
            if all(
                child.bounds[0] == children[0].bounds[0] for child in children
            ) and all(child.bounds[1] == children[0].bounds[1] for child in children):
                self.bounds = children[0].bounds
            else:
                raise ValueError("Cannot concatenate symbols with different bounds")
        super().__init__(*children, name=name)

        print_name = intersect(children[0]._raw_print_name, children[1]._raw_print_name)
        for child in children[2:]:
            print_name = intersect(print_name, child._raw_print_name)
        if print_name.endswith("_"):
            print_name = print_name[:-1]

        self.print_name = print_name


def substrings(s: str):
    for i in range(len(s)):
        for j in range(i, len(s)):
            yield s[i : j + 1]


def intersect(s1: str, s2: str):
    # find all the common strings between two strings
    all_intersects = set(substrings(s1)) & set(substrings(s2))
    # intersect is the longest such intercept
    if len(all_intersects) == 0:
        return ""
    intersect = max(all_intersects, key=len)
    # remove leading and trailing white space
    return intersect.lstrip().rstrip()


def simplified_concatenation(*children, name: str | None = None):
    """Perform simplifications on a concatenation."""
    # remove children that are None
    children = list(filter(lambda x: x is not None, children))
    # Simplify concatenation of broadcasts all with the same child to a single
    # broadcast across all domains
    if len(children) == 0:
        raise ValueError("Cannot create empty concatenation")
    elif len(children) == 1:
        return children[0]
    elif all(isinstance(child, pybamm.Variable) for child in children):
        return pybamm.ConcatenationVariable(*children, name=name)
    else:
        # Create Concatenation to easily read domains
        concat = Concatenation(*children, name=name)
        if all(
            isinstance(child, pybamm.Broadcast) and child.child == children[0].child
            for child in children
        ):
            unique_child = children[0].orphans[0]
            if isinstance(children[0], pybamm.PrimaryBroadcast):
                return pybamm.PrimaryBroadcast(unique_child, concat.domain, name=name)
            else:
                return pybamm.FullBroadcast(
                    unique_child, broadcast_domains=concat.domains, name=name
                )
        else:
            return concat


def concatenation(*children, name: str | None = None):
    """Helper function to create concatenations."""
    # TODO: add option to turn off simplifications
    return simplified_concatenation(*children, name=name)


def simplified_numpy_concatenation(*children):
    """Perform simplifications on a numpy concatenation."""
    # Turn a concatenation of concatenations into a single concatenation
    new_children = []
    for child in children:
        # extract any children from numpy concatenation
        if isinstance(child, NumpyConcatenation):
            new_children.extend(child.orphans)
        else:
            new_children.append(child)
    return pybamm.simplify_if_constant(NumpyConcatenation(*new_children))


def numpy_concatenation(*children):
    """Helper function to create numpy concatenations."""
    # TODO: add option to turn off simplifications
    return simplified_numpy_concatenation(*children)


def simplified_domain_concatenation(
    children: list[pybamm.Symbol],
    mesh: pybamm.Mesh,
    copy_this: DomainConcatenation | None = None,
):
    """Perform simplifications on a domain concatenation."""
    # Create the DomainConcatenation to read domain and child domain
    concat = DomainConcatenation(children, mesh, copy_this=copy_this)
    # Simplify Concatenation of StateVectors to a single StateVector
    # The sum of the evalation arrays of the StateVectors must be exactly 1
    if all(isinstance(child, pybamm.StateVector) for child in children):
        sv_children: list[pybamm.StateVector] = children  # type: ignore[assignment]
        longest_eval_array = len(sv_children[-1]._evaluation_array)
        eval_arrays = {}
        for child in sv_children:
            eval_arrays[child] = np.concatenate(
                [
                    child.evaluation_array,
                    np.zeros(longest_eval_array - len(child.evaluation_array)),
                ]
            )
        first_start = sv_children[0].y_slices[0].start
        last_stop = sv_children[-1].y_slices[-1].stop
        if all(
            sum(array for array in eval_arrays.values())[first_start:last_stop] == 1
        ):
            return pybamm.StateVector(
                slice(first_start, last_stop), domains=concat.domains
            )

    return pybamm.simplify_if_constant(concat)


def domain_concatenation(children: list[pybamm.Symbol], mesh: pybamm.Mesh):
    """Helper function to create domain concatenations."""
    # TODO: add option to turn off simplifications
    return simplified_domain_concatenation(children, mesh)



import json
import os

import numpy as np


def _process_name(name, path, ext):
    if not name.endswith(ext):
        name = name + ext

    # Set the path
    if path is not None:
        # first look in the specified path
        filename = os.path.join(path, name)
        if not os.path.exists(filename):
            # then look in the "data" subfolder
            filename = os.path.join(path, "data", name)
            if not os.path.exists(filename):
                raise FileNotFoundError(
                    "Could not find file '{}' in '{}' or '{}'".format(
                        name, path, os.path.join(path, "data")
                    )
                )
    else:
        filename = name
        # Split the name (in case original name was a path)
        _, name = os.path.split(filename)

    # Remove the extension from the name
    return (filename, name.split(".")[0])


def process_1D_data(name, path=None):
    """
    Process 1D data from a csv file
    """
    filename, name = _process_name(name, path, ".csv")

    data = np.genfromtxt(filename, delimiter=",", skip_header=1)
    x = data[:, 0]
    y = data[:, 1]

    # Save name and data
    return (name, ([x], y))


def process_2D_data(name, path=None):
    """
    Process 2D data from a JSON file
    """
    filename, name = _process_name(name, path, ".json")

    with open(filename) as jsonfile:
        json_data = json.load(jsonfile)
    data = json_data["data"]
    data[0] = [np.array(el) for el in data[0]]
    data[1] = np.array(data[1])
    return (name, tuple(data))


def process_2D_data_csv(name, path=None):
    """
    Process 2D data from a csv file. Assumes
    data is in the form of a three columns
    and that all data points lie on a regular
    grid. The first column is assumed to
    be the 'slowest' changing variable and
    the second column the 'fastest' changing
    variable, which is the C convention for
    indexing multidimensional arrays (as opposed
    to the Fortran convention where the 'fastest'
    changing variable comes first).

    Parameters
    ----------
    name : str
        The name to be given to the function
    path : str
        The path to the file where the three
        dimensional data is stored.

    Returns
    -------
    formatted_data: tuple
        A tuple containing the name of the function
        and the data formatted correctly for use
        within three-dimensional interpolants.
    """

    filename, name = _process_name(name, path, ".csv")

    data = np.genfromtxt(filename, delimiter=",", skip_header=1)

    x1 = np.unique(data[:, 0])
    x2 = np.unique(data[:, 1])

    value = data[:, 2]

    x = (x1, x2)

    value_data = value.reshape(len(x1), len(x2), order="C")

    formatted_data = (name, (x, value_data))

    return formatted_data


def process_3D_data_csv(name, path=None):
    """
    Process 3D data from a csv file. Assumes
    data is in the form of four columns and
    that all data points lie on a
    regular grid. The first column is assumed to
    be the 'slowest' changing variable and
    the third column the 'fastest' changing
    variable, which is the C convention for
    indexing multidimensional arrays (as opposed
    to the Fortran convention where the 'fastest'
    changing variable comes first).

    Parameters
    ----------
    name : str
        The name to be given to the function
    path : str
        The path to the file where the three
        dimensional data is stored.

    Returns
    -------
    formatted_data: tuple
        A tuple containing the name of the function
        and the data formatted correctly for use
        within three-dimensional interpolants.
    """

    filename, name = _process_name(name, path, ".csv")

    data = np.genfromtxt(filename, delimiter=",", skip_header=1)

    x1 = np.unique(data[:, 0])
    x2 = np.unique(data[:, 1])
    x3 = np.unique(data[:, 2])

    value = data[:, 3]

    x = (x1, x2, x3)

    value_data = value.reshape(len(x1), len(x2), len(x3), order="C")

    formatted_data = (name, (x, value_data))

    return formatted_data



#
# Basic lead-acid model
#
import pybamm

from .base_lead_acid_model import BaseModel


class BasicFull(BaseModel):
    """
    Porous electrode model for lead-acid, from :footcite:t:`Sulzer2019asymptotic`.

    This class differs from the :class:`pybamm.lead_acid.Full` model class in that it
    shows the whole model in a single class. This comes at the cost of flexibility in
    comparing different physical effects, and in general the main DFN class should be
    used instead.

    Parameters
    ----------
    name : str, optional
        The name of the model.

    """

    def __init__(self, name="Basic full model"):
        super().__init__({}, name)
        # `param` is a class containing all the relevant parameters and functions for
        # this model. These are purely symbolic at this stage, and will be set by the
        # `ParameterValues` class when the model is processed.

        ######################
        # Variables
        ######################
        # Variables that depend on time only are created without a domain
        Q = pybamm.Variable("Discharge capacity [A.h]")
        # Variables that vary spatially are created with a domain
        c_e_n = pybamm.Variable(
            "Negative electrolyte concentration [mol.m-3]",
            domain="negative electrode",
            scale=self.param.c_e_init,
        )
        c_e_s = pybamm.Variable(
            "Separator electrolyte concentration [mol.m-3]",
            domain="separator",
            scale=self.param.c_e_init,
        )
        c_e_p = pybamm.Variable(
            "Positive electrolyte concentration [mol.m-3]",
            domain="positive electrode",
            scale=self.param.c_e_init,
        )
        # Concatenations combine several variables into a single variable, to simplify
        # implementing equations that hold over several domains
        c_e = pybamm.concatenation(c_e_n, c_e_s, c_e_p)

        # Electrolyte potential
        phi_e_n = pybamm.Variable(
            "Negative electrolyte potential [V]",
            domain="negative electrode",
            reference=-self.param.n.prim.U_init,
        )
        phi_e_s = pybamm.Variable(
            "Separator electrolyte potential [V]",
            domain="separator",
            reference=-self.param.n.prim.U_init,
        )
        phi_e_p = pybamm.Variable(
            "Positive electrolyte potential [V]",
            domain="positive electrode",
            reference=-self.param.n.prim.U_init,
        )
        phi_e = pybamm.concatenation(phi_e_n, phi_e_s, phi_e_p)

        # Electrode potential
        phi_s_n = pybamm.Variable(
            "Negative electrode potential [V]", domain="negative electrode"
        )
        phi_s_p = pybamm.Variable(
            "Positive electrode potential [V]",
            domain="positive electrode",
            reference=self.param.ocv_init,
        )

        # Porosity
        eps_n = pybamm.Variable(
            "Negative electrode porosity", domain="negative electrode"
        )
        eps_s = pybamm.Variable("Separator porosity", domain="separator")
        eps_p = pybamm.Variable(
            "Positive electrode porosity", domain="positive electrode"
        )
        eps = pybamm.concatenation(eps_n, eps_s, eps_p)

        # Constant temperature
        T = self.param.T_init

        ######################
        # Other set-up
        ######################

        # Current density
        i_cell = self.param.current_density_with_time

        # transport_efficiency
        tor = pybamm.concatenation(
            eps_n**self.param.n.b_e, eps_s**self.param.s.b_e, eps_p**self.param.p.b_e
        )

        # Interfacial reactions
        F_RT = self.param.F / (self.param.R * T)
        Feta_RT_n = F_RT * (phi_s_n - phi_e_n - self.param.n.prim.U(c_e_n, T))
        j0_n = self.param.n.prim.j0(c_e_n, T)
        j_n = 2 * j0_n * pybamm.sinh(self.param.n.prim.ne / 2 * Feta_RT_n)
        j_s = pybamm.PrimaryBroadcast(0, "separator")
        Feta_RT_p = F_RT * (phi_s_p - phi_e_p - self.param.p.prim.U(c_e_p, T))
        j0_p = self.param.p.prim.j0(c_e_p, T)
        j_p = 2 * j0_p * pybamm.sinh(self.param.p.prim.ne / 2 * (Feta_RT_p))

        a_n = pybamm.Parameter("Negative electrode surface area to volume ratio [m-1]")
        a_p = pybamm.Parameter("Positive electrode surface area to volume ratio [m-1]")
        a_j_n = a_n * j_n
        a_j_p = a_p * j_p
        a_j = pybamm.concatenation(a_j_n, j_s, a_j_p)

        ######################
        # State of Charge
        ######################
        I = self.param.current_with_time
        # The `rhs` dictionary contains differential equations, with the key being the
        # variable in the d/dt
        self.rhs[Q] = I / 3600
        # Initial conditions must be provided for the ODEs
        self.initial_conditions[Q] = pybamm.Scalar(0)

        ######################
        # Current in the electrolyte
        ######################
        i_e = (self.param.kappa_e(c_e, T) * tor) * (
            self.param.chiRT_over_Fc(c_e, T) * pybamm.grad(c_e) - pybamm.grad(phi_e)
        )
        # multiply by Lx**2 to improve conditioning
        self.algebraic[phi_e] = (pybamm.div(i_e) - a_j) * self.param.L_x**2
        self.boundary_conditions[phi_e] = {
            "left": (pybamm.Scalar(0), "Neumann"),
            "right": (pybamm.Scalar(0), "Neumann"),
        }
        self.initial_conditions[phi_e] = -self.param.n.prim.U_init

        ######################
        # Current in the solid
        ######################
        i_s_n = (
            -self.param.n.sigma(T)
            * (1 - eps_n) ** self.param.n.b_s
            * pybamm.grad(phi_s_n)
        )
        sigma_eff_p = self.param.p.sigma(T) * (1 - eps_p) ** self.param.p.b_s
        i_s_p = -sigma_eff_p * pybamm.grad(phi_s_p)
        # The `algebraic` dictionary contains differential equations, with the key being
        # the main scalar variable of interest in the equation
        # multiply by Lx**2 to improve conditioning
        self.algebraic[phi_s_n] = (pybamm.div(i_s_n) + a_j_n) * self.param.L_x**2
        self.algebraic[phi_s_p] = (pybamm.div(i_s_p) + a_j_p) * self.param.L_x**2
        self.boundary_conditions[phi_s_n] = {
            "left": (pybamm.Scalar(0), "Dirichlet"),
            "right": (pybamm.Scalar(0), "Neumann"),
        }
        self.boundary_conditions[phi_s_p] = {
            "left": (pybamm.Scalar(0), "Neumann"),
            "right": (i_cell / pybamm.boundary_value(-sigma_eff_p, "right"), "Neumann"),
        }
        # Initial conditions must also be provided for algebraic equations, as an
        # initial guess for a root-finding algorithm which calculates consistent initial
        # conditions
        self.initial_conditions[phi_s_n] = pybamm.Scalar(0)
        self.initial_conditions[phi_s_p] = self.param.ocv_init

        ######################
        # Porosity
        ######################
        DeltaVsurf = pybamm.concatenation(
            pybamm.PrimaryBroadcast(self.param.n.DeltaVsurf, "negative electrode"),
            pybamm.PrimaryBroadcast(0, "separator"),
            pybamm.PrimaryBroadcast(self.param.p.DeltaVsurf, "positive electrode"),
        )
        deps_dt = DeltaVsurf * a_j / self.param.F
        self.rhs[eps] = deps_dt
        self.initial_conditions[eps] = self.param.epsilon_init
        self.events.extend(
            [
                pybamm.Event(
                    "Zero negative electrode porosity cut-off", pybamm.min(eps_n)
                ),
                pybamm.Event(
                    "Max negative electrode porosity cut-off", 1 - pybamm.max(eps_n)
                ),
                pybamm.Event(
                    "Zero positive electrode porosity cut-off", pybamm.min(eps_p)
                ),
                pybamm.Event(
                    "Max positive electrode porosity cut-off", 1 - pybamm.max(eps_p)
                ),
            ]
        )

        ######################
        # Electrolyte concentration
        ######################
        N_e = (
            -tor * self.param.D_e(c_e, T) * pybamm.grad(c_e)
            + self.param.t_plus(c_e, T) * i_e / self.param.F
        )
        s = pybamm.concatenation(
            pybamm.PrimaryBroadcast(self.param.n.prim.s_plus_S, "negative electrode"),
            pybamm.PrimaryBroadcast(0, "separator"),
            pybamm.PrimaryBroadcast(self.param.p.prim.s_plus_S, "positive electrode"),
        )
        self.rhs[c_e] = (1 / eps) * (
            -pybamm.div(N_e) + s * a_j / self.param.F - c_e * deps_dt
        )
        self.boundary_conditions[c_e] = {
            "left": (pybamm.Scalar(0), "Neumann"),
            "right": (pybamm.Scalar(0), "Neumann"),
        }
        self.initial_conditions[c_e] = self.param.c_e_init
        self.events.append(
            pybamm.Event(
                "Zero electrolyte concentration cut-off", pybamm.min(c_e) - 0.002
            )
        )

        ######################
        # (Some) variables
        ######################
        voltage = pybamm.boundary_value(phi_s_p, "right")
        # The `variables` dictionary contains all variables that might be useful for
        # visualising the solution of the model
        self.variables = {
            "Electrolyte concentration [mol.m-3]": c_e,
            "Current [A]": I,
            "Negative electrode potential [V]": phi_s_n,
            "Electrolyte potential [V]": phi_e,
            "Positive electrode potential [V]": phi_s_p,
            "Voltage [V]": voltage,
            "Porosity": eps,
        }
        self.events.extend(
            [
                pybamm.Event(
                    "Minimum voltage [V]", voltage - self.param.voltage_low_cut
                ),
                pybamm.Event(
                    "Maximum voltage [V]", self.param.voltage_high_cut - voltage
                ),
            ]
        )



#
# Geometric Parameters
#
import pybamm

from .base_parameters import BaseParameters


class GeometricParameters(BaseParameters):
    """
    Standard geometric parameters
    """

    def __init__(self, options=None):
        self.options = options
        self.n = DomainGeometricParameters("negative", self)
        self.s = DomainGeometricParameters("separator", self)
        self.p = DomainGeometricParameters("positive", self)
        self.domain_params = {
            "negative": self.n,
            "separator": self.s,
            "positive": self.p,
        }

        # Set parameters and scales
        self._set_parameters()

    def _set_parameters(self):
        """Defines the dimensional parameters."""
        for domain in self.domain_params.values():
            domain._set_parameters()

        # Macroscale geometry
        self.L_x = (
            self.n.L + self.s.L + self.p.L
        )  # Total distance between current collectors
        self.L = self.n.L_cc + self.L_x + self.p.L_cc  # Total cell thickness
        self.L_Li = pybamm.Parameter("Lithium counter electrode thickness [m]")
        self.L_y = pybamm.Parameter(
            "Electrode width [m]"
        )  # For a cylindrical cell L_y is the "unwound" length of the electrode
        self.L_z = pybamm.Parameter("Electrode height [m]")
        self.r_inner = pybamm.Parameter("Inner cell radius [m]")
        self.r_outer = pybamm.Parameter("Outer cell radius [m]")
        self.n_electrodes_parallel = pybamm.Parameter(
            "Number of electrodes connected in parallel to make a cell"
        )
        self.A_cc = (
            self.L_y * self.L_z * self.n_electrodes_parallel
        )  # Current collector cross sectional area

        # Cell surface area and volume (for thermal models only)
        cell_geometry = self.options.get("cell geometry", None)
        if cell_geometry == "pouch":
            # assuming a single-layer pouch cell for now, see
            # https://github.com/pybamm-team/PyBaMM/issues/1777
            self.A_cooling = 2 * (
                self.L_y * self.L_z + self.L_z * self.L + self.L_y * self.L
            )
            self.V_cell = self.L_y * self.L_z * self.L
        else:
            self.A_cooling = pybamm.Parameter("Cell cooling surface area [m2]")
            self.V_cell = pybamm.Parameter("Cell volume [m3]")


class DomainGeometricParameters(BaseParameters):
    def __init__(self, domain, main_param):
        self.domain = domain
        self.main_param = main_param

        if self.domain != "separator":
            self.prim = ParticleGeometricParameters(domain, "primary", main_param)
            self.sec = ParticleGeometricParameters(domain, "secondary", main_param)
            self.phase_params = {"primary": self.prim, "secondary": self.sec}
        else:
            self.phase_params = {}

    def _set_parameters(self):
        """Defines the dimensional parameters."""
        for phase in self.phase_params.values():
            phase._set_parameters()

        if self.domain == "separator":
            self.L = pybamm.Parameter("Separator thickness [m]")
            self.b_e = pybamm.Parameter("Separator Bruggeman coefficient (electrolyte)")
            self.tau_e = pybamm.Parameter("Separator tortuosity factor (electrolyte)")
            return

        Domain = self.domain.capitalize()

        # Macroscale geometry
        self.L_cc = pybamm.Parameter(f"{Domain} current collector thickness [m]")
        self.L = pybamm.Parameter(f"{Domain} electrode thickness [m]")

        # Tab geometry (for pouch cells)
        self.L_tab = pybamm.Parameter(f"{Domain} tab width [m]")
        self.centre_y_tab = pybamm.Parameter(f"{Domain} tab centre y-coordinate [m]")
        self.centre_z_tab = pybamm.Parameter(f"{Domain} tab centre z-coordinate [m]")
        self.A_tab = self.L_tab * self.L_cc  # Area of tab

        # Microscale geometry
        self.b_e = pybamm.Parameter(
            f"{Domain} electrode Bruggeman coefficient (electrolyte)"
        )
        self.b_s = pybamm.Parameter(
            f"{Domain} electrode Bruggeman coefficient (electrode)"
        )
        self.tau_e = pybamm.Parameter(
            f"{Domain} electrode tortuosity factor (electrolyte)"
        )
        self.tau_s = pybamm.Parameter(
            f"{Domain} electrode tortuosity factor (electrode)"
        )


class ParticleGeometricParameters(BaseParameters):
    def __init__(self, domain, phase, main_param):
        self.domain = domain
        self.phase = phase
        self.main_param = main_param
        self.set_phase_name()

    def _set_parameters(self):
        """Defines the dimensional parameters."""
        Domain = self.domain.capitalize()
        pref = self.phase_prefactor

        # Microscale geometry
        # Note: for li-ion cells, the definition of the surface area to
        # volume ratio is overwritten in lithium_ion_parameters.py to be computed
        # based on the assumed particle shape
        self.a = pybamm.Parameter(
            f"{pref}{Domain} electrode surface area to volume ratio [m-1]"
        )

        # Particle-size distribution geometry
        self.R_min = pybamm.Parameter(f"{pref}{Domain} minimum particle radius [m]")
        self.R_max = pybamm.Parameter(f"{pref}{Domain} maximum particle radius [m]")

    @property
    def R_typ(self):
        # evaluate the typical particle radius
        # in the middle of the electrode
        main = self.main_param
        if self.domain == "negative":
            x = main.n.L / 2
        elif self.domain == "positive":
            x = main.n.L + main.s.L + main.p.L / 2

        inputs = {"Through-cell distance (x) [m]": x}
        Domain = self.domain.capitalize()
        return pybamm.FunctionParameter(
            f"{self.phase_prefactor}{Domain} particle radius [m]", inputs
        )

    @property
    def R(self):
        if self.domain == "negative":
            x = pybamm.standard_spatial_vars.x_n
        elif self.domain == "positive":
            x = pybamm.standard_spatial_vars.x_p

        inputs = {"Through-cell distance (x) [m]": x}
        Domain = self.domain.capitalize()
        return pybamm.FunctionParameter(
            f"{self.phase_prefactor}{Domain} particle radius [m]", inputs
        )

    def f_a_dist(self, R):
        """
        Dimensional electrode area-weighted particle-size distribution
        """
        Domain = self.domain.capitalize()
        if len(R.domain) == 0:
            self.phase_prefactor = ""
        elif "primary" in R.domains["primary"][0]:
            self.phase_prefactor = "Primary: "
        elif "secondary" in R.domains["primary"][0]:
            self.phase_prefactor = "Secondary: "
        else:
            self.phase_prefactor = ""
        inputs = {f"{self.phase_prefactor}{Domain} particle-size variable [m]": R}
        return pybamm.FunctionParameter(
            f"{self.phase_prefactor}{Domain} "
            "area-weighted particle-size distribution [m-1]",
            inputs,
        )


geometric_parameters = GeometricParameters()



from typing import Any

import numpy as np
import numpy.typing as npt

import pybamm


class DiscreteTimeData(pybamm.Interpolant):
    """
    A class for representing data that is only defined at discrete points in time.
    This is implemented as a 1D interpolant with the time points as the nodes.

    Parameters
    ----------

    time_points : :class:`numpy.ndarray`
        The time points at which the data is defined
    data : :class:`numpy.ndarray`
        The data to be interpolated
    name : str
        The name of the data

    """

    def __init__(
        self, time_points: npt.NDArray[np.float64], data: npt.NDArray[Any], name: str
    ):
        super().__init__(time_points, data, pybamm.t, name)

    def create_copy(self, new_children=None, perform_simplifications=True):
        """See :meth:`pybamm.Symbol.new_copy()`."""
        return pybamm.DiscreteTimeData(self.x[0], self.y, self.name)


class DiscreteTimeSum(pybamm.UnaryOperator):
    """
    A node in the expression tree representing a discrete time sum operator.

    .. math::
        \\sum_{i=0}^{N} f(y(t_i), t_i)

    where f is the expression given by the child, and the sum is over the discrete
    time points t_i. The set of time points is given by the :class:`pybamm.DiscreteTimeData` node,
    which must be somewhere in the expression tree given by the child. If the child
    does not contain a :class:`pybamm.DiscreteTimeData` node, then an error will be raised when
    the node is created. If the child contains multiple :class:`pybamm.DiscreteTimeData` nodes,
    an error will be raised when the node is created.


    Parameters
    ----------
    child: :class:`pybamm.Symbol`
        The symbol to be summed

    Attributes
    ----------
    data: :class:`pybamm.DiscreteTimeData`
        The discrete time data node in the child

    Raises
    ------
    :class:`pybamm.ModelError`
        If the child does not contain a :class:`pybamm.DiscreteTimeData` node, or if the child
        contains multiple :class:`pybamm.DiscreteTimeData` nodes.
    """

    def __init__(self, child: pybamm.Symbol):
        self.data = None
        for node in child.pre_order():
            if isinstance(node, DiscreteTimeData):
                # Check that there is exactly one DiscreteTimeData node in the child
                if self.data is not None:
                    raise pybamm.ModelError(
                        "DiscreteTimeSum can only have one DiscreteTimeData node in the child"
                    )
                self.data = node
        if self.data is None:
            raise pybamm.ModelError(
                "DiscreteTimeSum must contain a DiscreteTimeData node"
            )
        super().__init__("discrete time sum", child)

    @property
    def sum_values(self):
        return self.data.y

    @property
    def sum_times(self):
        return self.data.x[0]

    def _unary_evaluate(self, child):
        # return result of evaluating the child, we'll only implement the sum once the model is solved (in pybamm.ProcessedVariable)
        return child



import pybamm


class NoDiffusion(pybamm.BaseSubModel):
    """
    Without Diffusion element for
    equivalent circuits.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, options=None):
        super().__init__(param)
        self.model_options = options

    def get_coupled_variables(self, variables):
        z = pybamm.PrimaryBroadcast(variables["SoC"], "ECMD particle")
        x = pybamm.SpatialVariable(
            "x ECMD", domain=["ECMD particle"], coord_sys="Cartesian"
        )
        z_surf = pybamm.surf(z)
        eta_diffusion = pybamm.Scalar(0)

        variables.update(
            {
                "Distributed SoC": z,
                "x ECMD": x,
                "Diffusion overpotential [V]": eta_diffusion,
                "Surface SoC": z_surf,
            }
        )

        return variables


class DiffusionElement(pybamm.BaseSubModel):
    """
    With Diffusion element for
    equivalent circuits.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, options=None):
        super().__init__(param)
        pybamm.citations.register("Fan2022")
        self.model_options = options

    def get_fundamental_variables(self):
        z = pybamm.Variable("Distributed SoC", domain="ECMD particle")
        x = pybamm.SpatialVariable(
            "x ECMD", domain=["ECMD particle"], coord_sys="Cartesian"
        )
        variables = {
            "Distributed SoC": z,
            "x ECMD": x,
        }
        return variables

    def get_coupled_variables(self, variables):
        z = variables["Distributed SoC"]
        soc = variables["SoC"]
        z_surf = pybamm.surf(z)
        eta_diffusion = -(self.param.ocv(z_surf) - self.param.ocv(soc))

        variables.update(
            {
                "Diffusion overpotential [V]": eta_diffusion,
                "Surface SoC": z_surf,
            }
        )

        return variables

    def set_rhs(self, variables):
        cell_capacity = self.param.cell_capacity
        current = variables["Current [A]"]
        z = variables["Distributed SoC"]

        # governing equations
        dzdt = pybamm.div(pybamm.grad(z)) / self.param.tau_D
        self.rhs = {z: dzdt}

        # boundary conditions
        lbc = pybamm.Scalar(0)
        rbc = -self.param.tau_D * current / (cell_capacity * 3600)
        self.boundary_conditions = {
            z: {"left": (lbc, "Neumann"), "right": (rbc, "Neumann")}
        }

    def set_initial_conditions(self, variables):
        z = variables["Distributed SoC"]
        self.initial_conditions = {z: self.param.initial_soc}

    def add_events_from(self, variables):
        z_surf = variables["Surface SoC"]
        self.events += [
            pybamm.Event("Minimum surface SoC", z_surf),
            pybamm.Event("Maximum surface SoC", 1 - z_surf),
        ]



#
# Class for full surface form electrolyte conductivity employing stefan-maxwell
#
import pybamm
from pybamm.models.submodels.electrolyte_conductivity.base_electrolyte_conductivity import (
    BaseElectrolyteConductivity,
)


class BaseModel(BaseElectrolyteConductivity):
    """Base class for conservation of charge in the electrolyte employing the
    Stefan-Maxwell constitutive equations employing the surface potential difference
    formulation.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain in which the model holds
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, domain, options=None):
        super().__init__(param, domain, options)

    def get_fundamental_variables(self):
        if self.domain == "separator":
            return {}

        domain, Domain = self.domain_Domain
        delta_phi = pybamm.Variable(
            f"{Domain} electrode surface potential difference [V]",
            domain=f"{domain} electrode",
            auxiliary_domains={"secondary": "current collector"},
            reference=self.domain_param.prim.U_init,
        )
        variables = self._get_standard_average_surface_potential_difference_variables(
            pybamm.x_average(delta_phi)
        )
        variables.update(
            self._get_standard_surface_potential_difference_variables(delta_phi)
        )

        return variables

    def get_coupled_variables(self, variables):
        Domain = self.domain.capitalize()

        if self.domain in ["negative", "positive"]:
            conductivity, sigma_eff = self._get_conductivities(variables)
            i_boundary_cc = variables["Current collector current density [A.m-2]"]
            c_e = variables[f"{Domain} electrolyte concentration [mol.m-3]"]
            delta_phi = variables[
                f"{Domain} electrode surface potential difference [V]"
            ]
            T = variables[f"{Domain} electrode temperature [K]"]

            i_e = conductivity * (
                self.param.chiRT_over_Fc(c_e, T) * pybamm.grad(c_e)
                + pybamm.grad(delta_phi)
                + i_boundary_cc / sigma_eff
            )
            variables[f"{Domain} electrolyte current density [A.m-2]"] = i_e

            phi_s = variables[f"{Domain} electrode potential [V]"]
            phi_e = phi_s - delta_phi

        elif self.domain == "separator":
            x_s = pybamm.standard_spatial_vars.x_s

            i_boundary_cc = variables["Current collector current density [A.m-2]"]
            c_e_s = variables["Separator electrolyte concentration [mol.m-3]"]
            if self.options.electrode_types["negative"] == "planar":
                phi_e_n_s = variables[
                    "Lithium metal interface electrolyte potential [V]"
                ]
            else:
                phi_e_n = variables["Negative electrolyte potential [V]"]
                phi_e_n_s = pybamm.boundary_value(phi_e_n, "right")
            tor_s = variables["Separator electrolyte transport efficiency"]
            T = variables["Separator temperature [K]"]

            chiRT_over_Fc_e_s = self.param.chiRT_over_Fc(c_e_s, T)
            kappa_s_eff = self.param.kappa_e(c_e_s, T) * tor_s

            phi_e = phi_e_n_s + pybamm.IndefiniteIntegral(
                chiRT_over_Fc_e_s * pybamm.grad(c_e_s) - i_boundary_cc / kappa_s_eff,
                x_s,
            )

            i_e = pybamm.PrimaryBroadcastToEdges(i_boundary_cc, "separator")
            variables[f"{Domain} electrolyte current density [A.m-2]"] = i_e

            # Update boundary conditions (for indefinite integral)
            self.boundary_conditions[c_e_s] = {
                "left": (pybamm.boundary_gradient(c_e_s, "left"), "Neumann"),
                "right": (pybamm.boundary_gradient(c_e_s, "right"), "Neumann"),
            }

        variables[f"{Domain} electrolyte potential [V]"] = phi_e

        if self.domain == "positive":
            phi_e_dict = {}
            i_e_dict = {}
            for domain in self.options.whole_cell_domains:
                Domain = domain.capitalize().split()[0]
                phi_e_dict[domain] = variables[f"{Domain} electrolyte potential [V]"]
                i_e_dict[domain] = variables[
                    f"{Domain} electrolyte current density [A.m-2]"
                ]

            variables.update(self._get_standard_potential_variables(phi_e_dict))

            i_e = pybamm.concatenation(*i_e_dict.values())
            variables.update(self._get_standard_current_variables(i_e))
            variables.update(self._get_electrolyte_overpotentials(variables))

        # save boundary conditons as variables
        if self.domain == "negative":
            grad_c_e = pybamm.boundary_gradient(c_e, "right")
            grad_left = -i_boundary_cc * pybamm.boundary_value(1 / sigma_eff, "left")
            grad_right = (
                (i_boundary_cc / pybamm.boundary_value(conductivity, "right"))
                - pybamm.boundary_value(self.param.chiRT_over_Fc(c_e, T), "right")
                * grad_c_e
                - i_boundary_cc * pybamm.boundary_value(1 / sigma_eff, "right")
            )

        elif self.domain == "positive":
            grad_c_e = pybamm.boundary_gradient(c_e, "left")
            grad_left = (
                (i_boundary_cc / pybamm.boundary_value(conductivity, "left"))
                - pybamm.boundary_value(self.param.chiRT_over_Fc(c_e, T), "left")
                * grad_c_e
                - i_boundary_cc * pybamm.boundary_value(1 / sigma_eff, "left")
            )
            grad_right = -i_boundary_cc * pybamm.boundary_value(1 / sigma_eff, "right")

        if self.domain in ["negative", "positive"]:
            variables.update(
                {
                    f"{self.domain} grad(delta_phi) left": grad_left,
                    f"{self.domain} grad(delta_phi) right": grad_right,
                    f"{self.domain} grad(c_e) internal": grad_c_e,
                }
            )
        return variables

    def _get_conductivities(self, variables):
        Domain = self.domain.capitalize()

        tor_e = variables[f"{Domain} electrolyte transport efficiency"]
        tor_s = variables[f"{Domain} electrode transport efficiency"]
        c_e = variables[f"{Domain} electrolyte concentration [mol.m-3]"]
        T = variables[f"{Domain} electrode temperature [K]"]
        sigma = self.domain_param.sigma(T)

        kappa_eff = self.param.kappa_e(c_e, T) * tor_e
        sigma_eff = sigma * tor_s
        conductivity = kappa_eff / (1 + kappa_eff / sigma_eff)

        return conductivity, sigma_eff

    def set_initial_conditions(self, variables):
        Domain = self.domain.capitalize()

        if self.domain == "separator":
            return

        delta_phi_e = variables[f"{Domain} electrode surface potential difference [V]"]
        delta_phi_e_init = self.domain_param.prim.U_init

        self.initial_conditions = {delta_phi_e: delta_phi_e_init}

    def set_boundary_conditions(self, variables):
        Domain = self.domain.capitalize()

        if self.domain == "separator":
            return

        c_e = variables[f"{Domain} electrolyte concentration [mol.m-3]"]
        delta_phi = variables[f"{Domain} electrode surface potential difference [V]"]

        grad_left = variables[f"{self.domain} grad(delta_phi) left"]
        grad_right = variables[f"{self.domain} grad(delta_phi) right"]
        grad_c_e = variables[f"{self.domain} grad(c_e) internal"]

        lbc = (grad_left, "Neumann")
        rbc = (grad_right, "Neumann")
        if self.domain == "negative":
            lbc_c_e = (pybamm.Scalar(0), "Neumann")
            rbc_c_e = (grad_c_e, "Neumann")
        elif self.domain == "positive":
            lbc_c_e = (grad_c_e, "Neumann")
            rbc_c_e = (pybamm.Scalar(0), "Neumann")

        # TODO: check why we still need the boundary conditions for c_e, once we have
        # internal boundary conditions
        self.boundary_conditions = {
            delta_phi: {"left": lbc, "right": rbc},
            c_e: {"left": lbc_c_e, "right": rbc_c_e},
        }

        if self.domain == "negative":
            phi_e = variables["Electrolyte potential [V]"]
            self.boundary_conditions.update(
                {
                    phi_e: {
                        "left": (pybamm.Scalar(0), "Neumann"),
                        "right": (pybamm.Scalar(0), "Neumann"),
                    }
                }
            )


class FullAlgebraic(BaseModel):
    """Full model for conservation of charge in the electrolyte employing the
    Stefan-Maxwell constitutive equations. (Full refers to unreduced by
    asymptotic methods)

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, domain, options=None):
        super().__init__(param, domain, options)

    def set_algebraic(self, variables):
        domain, Domain = self.domain_Domain

        if self.domain == "separator":
            return

        delta_phi = variables[f"{Domain} electrode surface potential difference [V]"]
        i_e = variables[f"{Domain} electrolyte current density [A.m-2]"]

        # Variable summing all of the interfacial current densities
        sum_a_j = variables[
            f"Sum of {domain} electrode volumetric "
            "interfacial current densities [A.m-3]"
        ]

        self.algebraic[delta_phi] = (pybamm.div(i_e) - sum_a_j) / self.param.a_j_scale


class FullDifferential(BaseModel):
    """Full model for conservation of charge in the electrolyte employing the
    Stefan-Maxwell constitutive equations and where capacitance is present.
    (Full refers to unreduced by asymptotic methods)

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, domain, options=None):
        super().__init__(param, domain, options)

    def set_rhs(self, variables):
        if self.domain == "separator":
            return

        domain, Domain = self.domain_Domain

        T = variables[f"{Domain} electrode temperature [K]"]
        C_dl = self.domain_param.C_dl(T)

        delta_phi = variables[f"{Domain} electrode surface potential difference [V]"]
        i_e = variables[f"{Domain} electrolyte current density [A.m-2]"]

        sum_a_j = variables[
            f"Sum of {domain} electrode volumetric "
            "interfacial current densities [A.m-3]"
        ]
        a = variables[f"{Domain} electrode surface area to volume ratio [m-1]"]

        self.rhs[delta_phi] = 1 / (a * C_dl) * (pybamm.div(i_e) - sum_a_j)



#
# Documentation utilities for actions on PyBaMM classes
#
from inspect import getmro


def copy_parameter_doc_from_parent(cls):
    """
    Add parameters from the base class to the docstring (Sphinx
    documentation). This constructs a new docstring by concatenating with
    formatting the base class docstring with the derived class. The
    base class docstring is truncated at the **Parameters** section.

    Usage: as a decorator @copy_parameter_doc_from_parent on derived
    class definition.
    """
    base_cls = getmro(cls)[1]
    parameters_section = "".join(base_cls.__doc__.partition("Parameters")[1:])
    cls.__doc__ += f"\n\n    {parameters_section}"
    return cls


# also implemented as a Sphinx extension in docs/sphinx_extend_parent.py, but this is
# not automatic, so we need to use the decorator manually in the class definition
def doc_extend_parent(cls):
    """
    Add a link to the base class in the docstring (Sphinx documentation) via the
    **Extends** directive. Constructs a new docstring element by concatenating
    with formatting the method resolution order (MRO) of the derived class.

    Usage: as a decorator @doc_extend_parent on derived class definition.
    """
    base_cls_name = f"{getmro(cls)[1].__module__}.{getmro(cls)[1].__name__}"
    cls.__doc__ += f"\n\n    **Extends:** :class:`{base_cls_name}`\n    "
    return cls



#
# Method for plotting voltage components
#
import numpy as np

from pybamm.simulation import Simulation
from pybamm.solvers.solution import Solution
from pybamm.util import import_optional_dependency


def plot_voltage_components(
    input_data,
    ax=None,
    show_legend=True,
    split_by_electrode=False,
    electrode_phases=("primary", "primary"),
    show_plot=True,
    **kwargs_fill,
):
    """
    Generate a plot showing the component overpotentials that make up the voltage

    Parameters
    ----------
    input_data : :class:`pybamm.Solution` or :class:`pybamm.Simulation`
        Solution or Simulation object from which to extract voltage components.
    ax : matplotlib Axis, optional
        The axis on which to put the plot. If None, a new figure and axis is created.
    show_legend : bool, optional
        Whether to display the legend. Default is True
    split_by_electrode : bool, optional
        Whether to show the overpotentials for the negative and positive electrodes
        separately. Default is False.
    electrode_phases : (str, str), optional
        The phases for which to plot the anode and cathode overpotentials, respectively.
        Default is `("primary", "primary")`.
    show_plot : bool, optional
        Whether to show the plots. Default is True. Set to False if you want to
        only display the plot after plt.show() has been called.
    kwargs_fill
        Keyword arguments: :obj:`matplotlib.axes.Axes.fill_between`
    """
    # Check if the input is a Simulation and extract Solution
    if isinstance(input_data, Simulation):
        solution = input_data.solution
    elif isinstance(input_data, Solution):
        solution = input_data

    # Remove phase name if the electrode is not composite
    electrode_phases = list(electrode_phases)
    if isinstance(solution.all_models[0].options["particle phases"], str):
        number_of_phases = (
            solution.all_models[0].options["particle phases"],
            solution.all_models[0].options["particle phases"],
        )
    else:
        number_of_phases = solution.all_models[0].options["particle phases"]
    for i in [0, 1]:
        if number_of_phases[i] == "1":
            electrode_phases[i] = ""
        else:
            electrode_phases[i] += " "
    num_cells = (
        solution["Battery voltage [V]"].entries[0] / solution["Voltage [V]"].entries[0]
    )
    full_cell = solution.all_models[0].options["working electrode"] == "both"
    plt = import_optional_dependency("matplotlib.pyplot")

    # Set a default value for alpha, the opacity
    kwargs_fill = {"alpha": 0.6, **kwargs_fill}

    if ax is not None:
        fig = None
        show_plot = False
    else:
        fig, ax = plt.subplots(figsize=(8, 4))

    if split_by_electrode is False:
        overpotentials = [
            "Battery particle concentration overpotential [V]",
            "X-averaged battery reaction overpotential [V]",
            "X-averaged battery concentration overpotential [V]",
            "X-averaged battery electrolyte ohmic losses [V]",
            "X-averaged battery solid phase ohmic losses [V]",
        ]
        labels = [
            "Particle concentration overpotential",
            "Reaction overpotential",
            "Electrolyte concentration overpotential",
            "Ohmic electrolyte overpotential",
            "Ohmic electrode overpotential",
        ]
    else:
        overpotentials = [
            f"Negative {electrode_phases[0]}particle concentration overpotential [V]",
            f"Positive {electrode_phases[1]}particle concentration overpotential [V]",
            f"X-averaged negative electrode {electrode_phases[0]}reaction overpotential [V]"
            if full_cell
            else "X-averaged battery negative reaction overpotential [V]",
            f"X-averaged positive electrode {electrode_phases[1]}reaction overpotential [V]",
            "X-averaged battery concentration overpotential [V]",
            "X-averaged battery electrolyte ohmic losses [V]",
            "X-averaged battery negative solid phase ohmic losses [V]",
            "X-averaged battery positive solid phase ohmic losses [V]",
        ]
        labels = [
            f"Negative particle {electrode_phases[0]}concentration overpotential",
            f"Positive particle {electrode_phases[1]}concentration overpotential",
            f"Negative {electrode_phases[0]}reaction overpotential",
            f"Positive {electrode_phases[1]}reaction overpotential",
            "Electrolyte concentration overpotential",
            "Ohmic electrolyte overpotential",
            "Ohmic negative electrode overpotential",
            "Ohmic positive electrode overpotential",
        ]
    # Only add the contact overpotential label if its values are not (numerically) all zero
    if not np.allclose(
        solution["Contact overpotential [V]"].entries, 0, atol=1e-12, equal_nan=True
    ):
        overpotentials.append("Contact overpotential [V]")
        labels.append("Contact overpotential")

    # Plot
    # Initialise
    time = solution["Time [h]"].entries
    if split_by_electrode is False:
        ocv = solution["Battery open-circuit voltage [V]"]
        initial_ocv = ocv(time[0])
        ocv = ocv.entries
        ax.fill_between(
            time, ocv, initial_ocv, **kwargs_fill, label="Open-circuit voltage"
        )
    else:
        ocp_n = solution[
            f"Negative electrode {electrode_phases[0]}bulk open-circuit potential [V]"
        ]
        ocp_p = solution[
            f"Positive electrode {electrode_phases[1]}bulk open-circuit potential [V]"
        ]
        initial_ocp_n = ocp_n(time[0]) * num_cells
        initial_ocp_p = ocp_p(time[0]) * num_cells
        initial_ocv = initial_ocp_p - initial_ocp_n
        delta_ocp_n = ocp_n.entries * num_cells - initial_ocp_n
        delta_ocp_p = ocp_p.entries * num_cells - initial_ocp_p
        ax.fill_between(
            time,
            initial_ocv - delta_ocp_n,
            initial_ocv,
            **kwargs_fill,
            label=f"Negative {electrode_phases[0]}open-circuit potential",
        )
        ax.fill_between(
            time,
            initial_ocv - delta_ocp_n + delta_ocp_p,
            initial_ocv - delta_ocp_n,
            **kwargs_fill,
            label=f"Positive {electrode_phases[1]}open-circuit potential",
        )
        ocv = initial_ocv - delta_ocp_n + delta_ocp_p
    top = ocv
    # Plot components
    for overpotential, label in zip(overpotentials, labels, strict=False):
        # negative overpotentials are positive for a discharge and negative for a charge
        # Contact overpotential is positive for a discharge and negative for a charge
        # so we have to multiply by -1 to show them correctly
        sgn = -1 if ("egative" in overpotential or "Contact" in overpotential) else 1
        multiplier = sgn if "attery" in overpotential else sgn * num_cells
        bottom = top + multiplier * solution[overpotential].entries
        ax.fill_between(time, bottom, top, **kwargs_fill, label=label)
        top = bottom

    V = solution["Battery voltage [V]"].entries
    ax.plot(time, V, "k--", label="Voltage")

    if show_legend:
        leg = ax.legend(loc="center left", bbox_to_anchor=(1.05, 0.5), frameon=True)
        leg.get_frame().set_edgecolor("k")
    if fig is not None:
        fig.tight_layout()

    # Labels
    ax.set_xlim([time[0], time[-1]])
    ax.set_xlabel("Time [h]")

    y_min, y_max = (
        0.98 * min(np.nanmin(V), np.nanmin(ocv)),
        1.02 * (max(np.nanmax(V), np.nanmax(ocv))),
    )
    ax.set_ylim([y_min, y_max])

    if show_plot:  # pragma: no cover
        plt.show()

    return fig, ax



import pybamm

from .base_lithium_ion_model import BaseModel


class Basic3DThermalSPM(BaseModel):
    """Single Particle Model (SPM) model of a lithium-ion battery, from
    :footcite:t:`Marquis2019`.

    This class differs from the :class:`pybamm.lithium_ion.SPM` model class in that it
    shows the whole model in a single class. This comes at the cost of flexibility in
    combining different physical effects, and in general the main SPM class should be
    used instead.

    The model is based on the SPM, but with a separate cell domain for the temperature
    variable. This allows for a more detailed treatment of the thermal effects in the
    cell, as the temperature can vary independently of the other variables in the
    model.

    Parameters
    ----------
    name : str, optional
        The name of the model.
    """

    def __init__(self, options=None, name="SPM with Separate Cell Domain"):
        super().__init__(options, name)
        pybamm.citations.register("Marquis2019")
        # `param` is a class containing all the relevant parameters and functions for
        # this model. These are purely symbolic at this stage, and will be set by the
        # `ParameterValues` class when the model is processed.

        ######################
        # Variables
        ######################
        # Variables that depend on time only are created without a domain
        Q = pybamm.Variable("Discharge capacity [A.h]")
        # Variables that vary spatially are created with a domain
        c_s_n = pybamm.Variable(
            "X-averaged negative particle concentration [mol.m-3]",
            domain="negative particle",
        )
        c_s_p = pybamm.Variable(
            "X-averaged positive particle concentration [mol.m-3]",
            domain="positive particle",
        )

        T = pybamm.Variable("Cell temperature [K]", domain="cell")

        if self.options.get("cell geometry") == "pouch":
            self.x_cell = pybamm.SpatialVariable("x", domain="cell")
            self.y_cell = pybamm.SpatialVariable("y", domain="cell")
            self.z_cell = pybamm.SpatialVariable("z", domain="cell")
            integration_vars = [self.x_cell, self.y_cell, self.z_cell]
        elif self.options.get("cell geometry") == "cylindrical":
            self.r_cell = pybamm.SpatialVariable("r_macro", domain="cell")
            self.z_cell = pybamm.SpatialVariable("z", domain="cell")
            integration_vars = [self.r_cell, self.z_cell]
        else:
            raise ValueError(
                f"Geometry type '{self.options.get('cell geometry')}' is not supported. "
                "Supported types are 'pouch' and 'cylindrical'."
            )  # pragma: no cover

        volume = pybamm.Integral(pybamm.PrimaryBroadcast(1.0, "cell"), integration_vars)
        T_av = pybamm.Integral(T, integration_vars) / volume

        ######################
        # Other set-up
        ######################

        # Current density
        i_cell = self.param.current_density_with_time
        a_n = 3 * self.param.n.prim.epsilon_s_av / self.param.n.prim.R_typ
        a_p = 3 * self.param.p.prim.epsilon_s_av / self.param.p.prim.R_typ
        j_n = i_cell / (self.param.n.L * a_n)
        j_p = -i_cell / (self.param.p.L * a_p)

        ######################
        # State of Charge
        ######################
        I = self.param.current_with_time
        # The `rhs` dictionary contains differential equations, with the key being the
        # variable in the d/dt
        self.rhs[Q] = I / 3600
        # Initial conditions must be provided for the ODEs
        self.initial_conditions[Q] = pybamm.Scalar(0)

        ######################
        # Particles
        ######################

        N_s_n = -self.param.n.prim.D(c_s_n, T_av) * pybamm.grad(c_s_n)
        N_s_p = -self.param.p.prim.D(c_s_p, T_av) * pybamm.grad(c_s_p)
        self.rhs[c_s_n] = -pybamm.div(N_s_n)
        self.rhs[c_s_p] = -pybamm.div(N_s_p)

        # Surf takes the surface value of a variable, i.e. its boundary value on the
        # right side. This is also accessible via `boundary_value(x, "right")`, with
        # "left" providing the boundary value of the left side
        c_s_surf_n = pybamm.surf(c_s_n)
        c_s_surf_p = pybamm.surf(c_s_p)

        # Boundary conditions must be provided for equations with spatial derivatives
        self.boundary_conditions[c_s_n] = {
            "left": (pybamm.Scalar(0), "Neumann"),
            "right": (
                -j_n / (self.param.F * pybamm.surf(self.param.n.prim.D(c_s_n, T_av))),
                "Neumann",
            ),
        }
        self.boundary_conditions[c_s_p] = {
            "left": (pybamm.Scalar(0), "Neumann"),
            "right": (
                -j_p / (self.param.F * pybamm.surf(self.param.p.prim.D(c_s_p, T_av))),
                "Neumann",
            ),
        }

        # c_n_init and c_p_init are functions of r and x, but for the SPM we
        # take the x-averaged value since there is no x-dependence in the particles
        self.initial_conditions[c_s_n] = pybamm.x_average(self.param.n.prim.c_init)
        self.initial_conditions[c_s_p] = pybamm.x_average(self.param.p.prim.c_init)
        # Events specify points at which a solution should terminate
        sto_surf_n = c_s_surf_n / self.param.n.prim.c_max
        sto_surf_p = c_s_surf_p / self.param.p.prim.c_max

        self.events += [
            pybamm.Event(
                "Minimum negative particle surface stoichiometry",
                pybamm.min(sto_surf_n) - 0.01,
            ),
            pybamm.Event(
                "Maximum negative particle surface stoichiometry",
                (1 - 0.01) - pybamm.max(sto_surf_n),
            ),
            pybamm.Event(
                "Minimum positive particle surface stoichiometry",
                pybamm.min(sto_surf_p) - 0.01,
            ),
            pybamm.Event(
                "Maximum positive particle surface stoichiometry",
                (1 - 0.01) - pybamm.max(sto_surf_p),
            ),
        ]

        # Note that the SPM does not have any algebraic equations, so the `algebraic`
        # dictionary remains empty

        ######################
        # (Some) variables
        ######################
        # Interfacial reactions
        RT_F = self.param.R * T_av / self.param.F
        j0_n = self.param.n.prim.j0(self.param.c_e_init_av, c_s_surf_n, T_av)
        j0_p = self.param.p.prim.j0(self.param.c_e_init_av, c_s_surf_p, T_av)
        eta_n = (2 / self.param.n.prim.ne) * RT_F * pybamm.arcsinh(j_n / (2 * j0_n))
        eta_p = (2 / self.param.p.prim.ne) * RT_F * pybamm.arcsinh(j_p / (2 * j0_p))
        phi_s_n = 0
        phi_e = -eta_n - self.param.n.prim.U(sto_surf_n, T_av)
        phi_s_p = eta_p + phi_e + self.param.p.prim.U(sto_surf_p, T_av)
        V = phi_s_p
        num_cells = pybamm.Parameter(
            "Number of cells connected in series to make a battery"
        )

        # Compute heat source

        # Reversible (entropic) heating
        dUdT_n = self.param.n.prim.dUdT(sto_surf_n)
        dUdT_p = self.param.p.prim.dUdT(sto_surf_p)
        Q_rev_n = a_n * j_n * T_av * dUdT_n
        Q_rev_p = a_p * j_p * T_av * dUdT_p

        # Irreversible (Joule) heating
        Q_irr_n = a_n * j_n * eta_n
        Q_irr_p = a_p * j_p * eta_p

        # Total heating in each electrode domain
        Q_total_n = Q_rev_n + Q_irr_n
        Q_total_p = Q_rev_p + Q_irr_p

        # Average the 1D heat source and broadcast to the 3D 'cell' domain
        L_n = self.param.n.L
        L_p = self.param.p.L
        L_x = self.param.L_x  # Total cell thickness

        # Calculate the true volume-weighted average heat source in W/m^3
        Q_vol_avg = (Q_total_n * L_n + Q_total_p * L_p) / L_x

        # Broadcast this uniform volumetric heat source to the entire 3D cell domain
        Q_source = pybamm.PrimaryBroadcast(Q_vol_avg, "cell")

        # Wrap in source term to get the correct mass matrix
        Q_source = pybamm.source(Q_source, T)

        # Define the 3D heat equation
        # Effective parameters are functions of temperature
        rho_c_p_eff = self.param.rho_c_p_eff(T)
        lambda_eff = self.param.lambda_eff(T)

        # The heat equation is d(rho*cp*T)/dt = div(lambda*grad(T)) + Q
        term1 = lambda_eff * pybamm.laplacian(T)
        term2 = pybamm.inner(pybamm.grad(lambda_eff), pybamm.grad(T))

        self.rhs[T] = (term1 + term2 + Q_source) / rho_c_p_eff

        # Cooling taken care of by boundary conditions

        self.initial_conditions[T] = pybamm.PrimaryBroadcast(self.param.T_init, "cell")
        self.set_thermal_bcs(T)

        whole_cell = ["negative electrode", "separator", "positive electrode"]
        # The `variables` dictionary contains all variables that might be useful for
        # visualising the solution of the model
        # Primary broadcasts are used to broadcast scalar quantities across a domain
        # into a vector of the right shape, for multiplying with other vectors
        self.variables = {
            "Time [s]": pybamm.t,
            "Discharge capacity [A.h]": Q,
            "X-averaged negative particle concentration [mol.m-3]": c_s_n,
            "Negative particle surface "
            "concentration [mol.m-3]": pybamm.PrimaryBroadcast(
                c_s_surf_n, "negative electrode"
            ),
            "Electrolyte concentration [mol.m-3]": pybamm.PrimaryBroadcast(
                self.param.c_e_init_av, whole_cell
            ),
            "X-averaged positive particle concentration [mol.m-3]": c_s_p,
            "Positive particle surface "
            "concentration [mol.m-3]": pybamm.PrimaryBroadcast(
                c_s_surf_p, "positive electrode"
            ),
            "Current [A]": I,
            "Current variable [A]": I,  # for compatibility with pybamm.Experiment
            "Negative electrode potential [V]": pybamm.PrimaryBroadcast(
                phi_s_n, "negative electrode"
            ),
            "Electrolyte potential [V]": pybamm.PrimaryBroadcast(phi_e, whole_cell),
            "Positive electrode potential [V]": pybamm.PrimaryBroadcast(
                phi_s_p, "positive electrode"
            ),
            "Voltage [V]": V,
            "Battery voltage [V]": V * num_cells,
        }

        # Add new thermal variables
        self.variables.update(
            {
                "Cell temperature [K]": T,
                "Volume-averaged cell temperature [K]": T_av,
            }
        )

        # Events specify points at which a solution should terminate
        self.events += [
            pybamm.Event("Minimum voltage [V]", V - self.param.voltage_low_cut),
            pybamm.Event("Maximum voltage [V]", self.param.voltage_high_cut - V),
        ]

    def set_thermal_bcs(self, T):
        geometry_type = self.options.get("cell geometry", "pouch")
        if geometry_type == "pouch":
            # Reuse the spatial variables created in __init__
            T_amb = self.param.T_amb(self.y_cell, self.z_cell, pybamm.t)
            face_params = {
                "x_min": self.param.h_edge_x_min,
                "x_max": self.param.h_edge_x_max,
                "y_min": self.param.h_edge_y_min,
                "y_max": self.param.h_edge_y_max,
                "z_min": self.param.h_edge_z_min,
                "z_max": self.param.h_edge_z_max,
            }
        elif geometry_type == "cylindrical":
            # Reuse the spatial variables created in __init__
            T_amb = self.param.T_amb(self.r_cell, self.z_cell, pybamm.t)
            face_params = {
                "r_min": self.param.h_edge_radial_min,
                "r_max": self.param.h_edge_radial_max,
                "z_min": self.param.h_edge_z_min,
                "z_max": self.param.h_edge_z_max,
            }
        else:
            raise ValueError(
                f"Geometry type '{geometry_type}' is not supported. "
                "Supported types are 'pouch' and 'cylindrical'."
            )  # pragma: no cover

        self.boundary_conditions[T] = {}

        for face, h_coeff in face_params.items():
            # Evaluate T and lambda_eff at the boundary
            T_boundary = pybamm.boundary_value(T, face)
            lambda_eff_boundary = pybamm.boundary_value(self.param.lambda_eff(T), face)
            # T_amb is already evaluated at the boundary coordinates
            q_face = -h_coeff * (T_boundary - T_amb)
            self.boundary_conditions[T][face] = (
                q_face / lambda_eff_boundary,
                "Neumann",
            )
